#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import unpack
from traceback import format_exc

from reclaimer import data_extraction
from reclaimer.util import get_is_xbox_map
from reclaimer.meta.halo_map import map_header_def, tag_index_pc_def
from reclaimer.meta.halo1_rsrc_map import lite_halo1_rsrc_map_def as halo1_rsrc_map_def
from reclaimer.meta.wrappers.byteswapping import raw_block_def, byteswap_pcm16_samples
from reclaimer.meta.wrappers.map_pointer_converter import MapPointerConverter
from reclaimer.meta.wrappers.tag_index_manager import TagIndexManager
from reclaimer.meta.wrappers.halo_map import HaloMap

from supyr_struct.buffer import BytearrayBuffer, get_rawdata
from supyr_struct.field_types import FieldType

# this is ultra hacky, but it seems to be the only
# way to fix the tagid for the sounds resource map
sound_rsrc_id_map = {
    92: 7,  # sound\sfx\impulse\impacts\smallrock
    93: 8,  # sound\sfx\impulse\impacts\medrocks
    94: 9,  # sound\sfx\impulse\impacts\lrgrocks

    125: 12,  # sound\sfx\impulse\impacts\metal_chips
    126: 13,  # sound\sfx\impulse\impacts\metal_chip_med

    372: 61,  # sound\sfx\impulse\shellcasings\double_shell_dirt
    373: 62,  # sound\sfx\impulse\shellcasings\multi_shell_dirt
    374: 63,  # sound\sfx\impulse\shellcasings\single_shell_metal
    375: 64,  # sound\sfx\impulse\shellcasings\double_shell_metal
    376: 65,  # sound\sfx\impulse\shellcasings\multi_shell_metal

    1545: 264,  # sound\sfx\impulse\glass\glass_medium
    1546: 265,  # sound\sfx\impulse\glass\glass_large
    }

DEFAULT_LOC_TAG_COUNT = 176
DEFAULT_SOUNDS_TAG_COUNT = 376
DEFAULT_BITMAPS_TAG_COUNT = 853


# Tag classes aren't stored in the cache maps, so we need to
# have a cache of them somewhere. Might as well do it manually
loc_exts = {0:'font', 1:'font', 4:'hud_message_text', 56:'font', 58:'font'}

bitmap_exts = ('bitmap', ) * DEFAULT_BITMAPS_TAG_COUNT
sound_exts  = ('sound', ) * DEFAULT_SOUNDS_TAG_COUNT
loc_exts    = tuple(loc_exts.get(i, 'unicode_string_list')
                    for i in range(DEFAULT_LOC_TAG_COUNT))


def inject_sound_data(map_data, rsrc_data, rawdata_ref, map_magic):
    if not rawdata_ref.size:
        rawdata_ref.data = b''
        return

    if rawdata_ref.flags.data_in_resource_map:
        data, ptr = rsrc_data, rawdata_ref.raw_pointer
    elif rawdata_ref.pointer == 0:
        data, ptr = map_data,  rawdata_ref.raw_pointer
    else:
        data, ptr = map_data,  rawdata_ref.pointer + map_magic

    data.seek(ptr)
    rawdata_ref.data = data.read(rawdata_ref.size)


class Halo1RsrcMap(HaloMap):
    '''Generation 1 resource map'''

    # the original resource map header
    rsrc_map = None

    tag_classes = None
    tag_headers = None

    tag_defs_module = "reclaimer.hek.defs"
    tag_classes_to_load = ("bitm", "snd!", "font", "hmt ", "ustr")

    data_extractors = data_extraction.h1_data_extractors

    def __init__(self, maps=None):
        HaloMap.__init__(self, maps)
        self.setup_tag_headers()

    def load_map(self, map_path, **kwargs):
        map_data = get_rawdata(filepath=map_path, writable=False)

        resource_type = unpack("<I", map_data.read(4))[0]
        map_data.seek(0)
        rsrc_map = self.rsrc_map = halo1_rsrc_map_def.build(rawdata=map_data)

        self.orig_tag_index = rsrc_map.data.tags

        # check if this is a pc or ce cache. cant rip pc ones
        pth = self.orig_tag_index[0].tag.path if self.orig_tag_index else ""
        self.filepath = map_path

        ce_engine = ""
        for halo_map in self.maps.values():
            ce_engine = getattr(halo_map, "engine")
            if ce_engine:
                break

        rsrc_tag_count = len(rsrc_map.data.tags)
        if resource_type == 3 or (pth.endswith('__pixels') or
                                  pth.endswith('__permutations')):
            if ce_engine:
                self.engine = ce_engine
            else:
                self.engine = "halo1ce"
        elif ((resource_type == 1 and rsrc_tag_count == 1107) or
              (resource_type == 2 and rsrc_tag_count == 7192)):
            self.engine = "halo1pcdemo"
        else:
            self.engine = "halo1pc"

        # so we don't have to redo a lot of code, we'll make a
        # fake tag_index and map_header and just fill in info
        self.map_header = head = map_header_def.build()
        self.tag_index  = tags = tag_index_pc_def.build()
        self.map_magic  = 0
        self.map_data   = map_data
        self.is_resource = True

        self.index_magic = 0

        index_mul = 2
        if self.engine == "halo1pc" or resource_type == 3:
            index_mul = 1

        rsrc_tag_count = rsrc_tag_count // index_mul
        if resource_type == 1:
            head.map_name = "bitmaps"
            tag_classes, def_cls = bitmap_exts, 'bitmap'
        elif resource_type == 2:
            head.map_name = "sounds"
            tag_classes, def_cls = sound_exts, 'sound'
        elif resource_type == 3:
            head.map_name = "loc"
            tag_classes, def_cls = loc_exts, 'NONE'
            if self.engine == "halo1ce" and rsrc_tag_count != 176:
                # this is a custom edition loc.map, but we can't trust it to be accurate
                # if it contains a different number than 176 tags.
                tag_classes = ()
        else:
            raise ValueError("Unknown resource map type.")

        # allow an override to be specified before the map is loaded
        if self.tag_classes is None:
            self.tag_classes = tag_classes

        self.maps[head.map_name] = self
        self.map_name = head.map_name
        self.tag_classes += (def_cls, )*(rsrc_tag_count - len(self.tag_classes))
        tags.tag_index.extend(rsrc_tag_count)

        tags.scenario_tag_id = 0
        tags.tag_count = rsrc_tag_count
        # fill in the fake tag_index
        for i in range(rsrc_tag_count):
            j = i*index_mul
            if index_mul != 1:
                j += 1

            tag_ref = tags.tag_index[i]
            tag_ref.class_1.set_to(self.tag_classes[i])
            tag_ref.id = i

            tag_ref.meta_offset  = rsrc_map.data.tags[j].offset
            tag_ref.path = rsrc_map.data.tags[j].tag.path
            # not necessary to set it as indexed, as it isn't used
            # when extracting the tags, and leaving it as False
            # allows efinery to display pointer information, and
            # change the classes of tags in custom loc maps.
            #tag_ref.indexed      = 1

        # apparently this is needed(thank spv3/open sauce v4/neil
        #                           for fucking up resource maps)
        self.basic_deprotection()

        self.map_pointer_converter = MapPointerConverter((0, 0, 0xFFffFFff))
        self.tag_index_manager = TagIndexManager(tags.tag_index)
        self.snd_rsrc_tag_index_manager = TagIndexManager(
            tags.tag_index, sound_rsrc_id_map)
        self.clear_map_cache()

    def get_dependencies(self, meta, tag_id, tag_cls):
        if tag_cls != "snd!":
            return ()

        tag_id = meta.promotion_sound.id & 0xFFff
        if tag_id not in range(len(self.tag_index.tag_index)):
            return ()

        return [self.tag_index.tag_index[tag_id]]

    def is_indexed(self, tag_id):
        return True

    def get_meta(self, tag_id, reextract=False, **kw):
        '''Returns just the meta of the tag without any raw data.'''
        if tag_id is None:
            return

        # if we are given a 32bit tag id, mask it off
        tag_id &= 0xFFFF
        tag_index_ref = self.tag_index_manager.get_tag_index_ref(tag_id)
        if tag_index_ref is None:
            return
        tag_cls = dict(
            sound="snd!", bitmap="bitm", font="font",
            unicode_string_list="ustr", hud_message_text="hmt ").get(
                tag_index_ref.class_1.enum_name)

        kwargs = dict(parsing_resource=True)
        desc = self.get_meta_descriptor(tag_cls)
        if desc is None or self.engine not in ("halo1ce", "halo1yelo",
                                               "halo1vap"):
            return
        elif tag_cls != 'snd!':
            # the pitch ranges pointer in resource sound tags is invalid, so
            # for sounds we treat it as if we're parsing a tag(pointerless).
            # only provide a pointer converter for bitmap and loc maps.
            kwargs['map_pointer_converter'] = self.map_pointer_converter

        block = [None]

        self.record_map_cache_read(tag_id, 0)  # cant get size quickly enough
        if self.map_cache_over_limit():
            self.clear_map_cache()

        try:
            FieldType.force_little()
            desc['TYPE'].parser(
                desc, parent=block, attr_index=0, rawdata=self.map_data,
                tag_index_manager=self.snd_rsrc_tag_index_manager,
                tag_cls=tag_cls, root_offset=tag_index_ref.meta_offset,
                indexed=True, **kwargs)
            FieldType.force_normal()

            if not kw.get("ignore_rawdata", False):
                self.inject_rawdata(block[0], tag_cls, tag_index_ref)
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                return

        return block[0]

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        magic      = self.map_magic
        engine     = self.engine
        map_data   = self.map_data
        tag_index  = self.tag_index
        is_xbox = get_is_xbox_map(engine)

        if tag_cls == "bitm":
            # set the size of the compressed plate data to nothing
            meta.compressed_color_plate_data.STEPTREE = BytearrayBuffer()

            new_pixels_offset = 0

            # uncheck the prefer_low_detail flag and
            # set up the pixels_offset correctly.
            for bitmap in meta.bitmaps.STEPTREE:
                bitmap.flags.prefer_low_detail = is_xbox
                bitmap.pixels_offset = new_pixels_offset
                new_pixels_offset += bitmap.pixels_meta_size

                # clear some meta-only fields
                bitmap.pixels_meta_size = 0
                bitmap.bitmap_id_unknown1 = bitmap.bitmap_id_unknown2 = 0
                bitmap.bitmap_data_pointer = bitmap.base_address = 0

        elif tag_cls == "snd!":
            meta.maximum_bend_per_second = meta.maximum_bend_per_second ** 30
            for pitch_range in meta.pitch_ranges.STEPTREE:
                for permutation in pitch_range.permutations.STEPTREE:
                    if permutation.compression.enum_name == "none":
                        # byteswap pcm audio
                        byteswap_pcm16_samples(permutation.samples)

        return meta

    def inject_rawdata(self, meta, tag_cls, tag_index_ref):
        bitmaps = self.maps.get("bitmaps")
        sounds  = self.maps.get("sounds")
        loc     = self.maps.get("loc")

        magic  = self.map_magic
        engine = self.engine

        map_data = self.map_data

        try:   bitmap_data = bitmaps.map_data
        except Exception:    bitmap_data = None
        try:   sound_data = sounds.map_data
        except Exception:   sound_data = None
        try:   loc_data = loc.map_data
        except Exception: loc_data = None

        is_not_indexed = not self.is_indexed(tag_index_ref.id & 0xFFff)
        might_be_in_rsrc = engine in ("halo1pc", "halo1pcdemo",
                                      "halo1ce", "halo1yelo", "halo1vap")
        might_be_in_rsrc &= not self.is_resource

        # get some rawdata that would be pretty annoying to do in the parser
        if tag_cls == "bitm":
            # grab bitmap data from map
            new_pixels = BytearrayBuffer()

            # to enable compatibility with my bitmap converter we'll set the
            # base address to a certain constant based on the console platform
            is_xbox = get_is_xbox_map(engine)
            for bitmap in meta.bitmaps.STEPTREE:
                pixel_data = map_data
                if might_be_in_rsrc and bitmap.flags.data_in_resource_map:
                    pixel_data = bitmap_data

                if pixel_data is None: return

                # grab the bitmap data from this map(no magic used)
                pixel_data.seek(bitmap.pixels_offset)
                new_pixels += pixel_data.read(bitmap.pixels_meta_size)

                bitmap.base_address = 1073751810 * is_xbox

            meta.processed_pixel_data.STEPTREE = new_pixels
        elif tag_cls == "font":
            # might need to grab pixel data from resource map
            meta_offset = tag_index_ref.meta_offset

            if is_not_indexed:
                return meta
            elif not self.is_resource:
                if loc is None or loc.map_header is None: return
                meta_offset = loc.rsrc_map.data.tags[meta_offset].tag.offset

            if loc_data is None: return

            loc_data.seek(meta.pixels.pointer + meta_offset)
            meta.pixels.data = loc_data.read(meta.pixels.size)
        elif tag_cls == "hmt ":
            # might need to grab string data from resource map
            meta_offset = tag_index_ref.meta_offset

            if is_not_indexed:
                return meta
            elif not self.is_resource:
                if loc is None or loc.map_header is None: return
                meta_offset = loc.rsrc_map.data.tags[meta_offset].tag.offset

            b = meta.string
            loc_data.seek(b.pointer + meta_offset)
            meta.string.data = loc_data.read(b.size).decode('utf-16-le')
        elif tag_cls == "snd!":
            # might need to get samples and permutations from the resource map
            is_pc = engine in ("halo1pc", "halo1pcdemo")
            is_ce = engine in ("halo1ce", "halo1yelo", "halo1vap")
            if not(is_pc or is_ce):
                return meta
            elif sound_data is None:
                return

            # ce tagpaths are in the format:  path__permutations
            #     ex: sound\sfx\impulse\coolant\enter_water__permutations
            #
            # pc tagpaths are in the format:  path__pitchrange__permutation
            #     ex: sound\sfx\impulse\coolant\enter_water__0__0
            other_data = map_data
            sound_magic = 0 - magic
            # DO NOT optimize this section. The logic is like this on purpose
            if is_pc:
                pass
            elif self.is_resource:
                other_data = sound_data
                sound_magic = tag_index_ref.meta_offset + meta.get_size()
            elif sounds is None:
                return

            for pitches in meta.pitch_ranges.STEPTREE:
                for perm in pitches.permutations.STEPTREE:
                    for b in (perm.samples, perm.mouth_data, perm.subtitle_data):
                        inject_sound_data(other_data, sound_data, b, sound_magic)

        elif tag_cls == "ustr":
            # might need to grab string data from resource map
            meta_offset = tag_index_ref.meta_offset

            if is_not_indexed:
                return meta
            elif not self.is_resource:
                if loc is None or loc.map_header is None: return
                meta_offset = loc.rsrc_map.data.tags[meta_offset].tag.offset

            string_blocks = meta.strings.STEPTREE

            if len(string_blocks):
                desc = string_blocks[0].get_desc('STEPTREE')
                parser = desc['TYPE'].parser

            try:
                FieldType.force_little()
                for b in string_blocks:
                    parser(desc, None, b, 'STEPTREE',
                           loc_data, meta_offset, b.pointer)
                FieldType.force_normal()
            except Exception:
                print(format_exc())
                FieldType.force_normal()
                raise

    def generate_map_info_string(self):
        if hasattr(self.map_data, '__len__'):
            map_size = str(len(self.map_data))
        elif (hasattr(self.map_data, 'seek') and
              hasattr(self.map_data, 'tell')):
            curr_pos = self.map_data.tell()
            self.map_data.seek(0, 2)
            map_size = str(self.map_data.tell())
            self.map_data.seek(curr_pos)
        else:
            map_size = "unknown"

        return """%s
General:
    engine         == %s
    file size      == %s
    resource type  == %s

Header:
    tag count         == %s
    tag paths pointer == %s
    tag index pointer == %s""" % (
        self.filepath, self.engine, map_size, self.map_name,
        self.tag_index.tag_count, self.rsrc_map.data.tag_paths_pointer,
        self.rsrc_map.data.tag_headers_pointer)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	Mapped out value types
# revision: 3		author: Lord Zedd
# 	Updated and copypasted.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


mdl3_camera_refinement_zoom_data_1 = Struct("zoom_data_1",
    h3_rawdata_ref("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


mdl3_camera_refinement_zoom_data_2 = Struct("zoom_data_2",
    h3_rawdata_ref("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


mdl3_camera_refinement = Struct("camera_refinement",
    h3_string_id("biped", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    h3_reflexive("zoom_data_1", mdl3_camera_refinement_zoom_data_1),
    h3_reflexive("zoom_data_2", mdl3_camera_refinement_zoom_data_2),
    VISIBLE=False,
    ENDIAN=">", SIZE=60
    )


mdl3_body = Struct("tagdata",
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_reflexive("camera_refinement", mdl3_camera_refinement),
    ENDIAN=">", SIZE=56
    )


def get():
    return mdl3_def

mdl3_def = TagDef("mdl3",
    h3_blam_header('mdl3'),
    mdl3_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["mdl3"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


wind_body = Struct("tagdata",
    h3_rawdata_ref("function_0"),
    h3_rawdata_ref("function_1"),
    h3_rawdata_ref("function_2"),
    h3_rawdata_ref("function_3"),
    h3_rawdata_ref("function_4"),
    Float("unknown"),
    h3_dependency("warp_bitmap"),
    ENDIAN=">", SIZE=120
    )


def get():
    return wind_def

wind_def = TagDef("wind",
    h3_blam_header('wind'),
    wind_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wind"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from traceback import format_exc
from reclaimer.enums import hmt_icon_types

__all__ = ("compile_hud_message_text",
           "compile_unicode_string_list", "compile_string_list")


icon_type_map = {hmt_icon_types[i]: i for i in range(len(hmt_icon_types))}
MAX_ICON_NAME_LENGTH = max(*(len(name) for name in hmt_icon_types))


def compile_hud_message_text(hmt_tag, hmt_string_data):
    message_strings = {}
    message_names = []

    line_num = 0
    for line in hmt_string_data.lstrip('\ufeff').split("\n"):
        line_num += 1
        if not line:
            continue
        elif "=" in line:
            name, message = line.split("=", 1)
        else:
            name, message = line, ""

        if name in message_strings:
            print("    ERROR: Duplicate message name on line %s" % line_num)
            continue

        if len(name) > 31:
            print(("    WARNING: Message name on line %s is too long. " +
                   "Truncating name to '%s'" ) % (line_num, name[: 31]))
            name = name[: 31]

        if name and message:
            message_strings[name] = message
            message_names.append(name)
        elif name and not message and "=" in line:
            print("    WARNING: Empty message on line %s" % line_num)
            message_strings[name] = ""
            message_names.append(name)
        elif not name and message:
            print("    ERROR: No message name on line %s" % line_num)
        else:
            print("    ERROR: No name or message on line %s" % line_num)

    if len(message_strings) > 1024:
        print("    ERROR: Too many hud messages. Please remove %s of them." %
              (len(message_strings) - 1024))
        return

    tagdata = hmt_tag.data.tagdata
    elements = tagdata.message_elements.STEPTREE
    messages = tagdata.messages.STEPTREE
    tagdata.string.data = ""
    del elements[:]
    del messages[:]

    error = False
    text_blob = ""
    for name in message_names:
        message_string = message_strings[name]

        messages.append()
        message = messages[-1]
        message.name = name
        message.text_start = len(text_blob)
        message.element_index = len(elements)

        if len(message_string) == 0:
            elements.append()
            element = elements[-1]
            element.data.set_active("text")
            element.data.text.data = 1  # set length to 1
            message.element_count = 1
            text_blob += "\x00"
            continue

        i = 0
        while i < len(message_string):
            elements.append()
            element = elements[-1]
            message.element_count += 1

            element_base = i
            icon_name = ""
            curr_text = ""
            c = message_string[i]
            if c == "%":
                element.type.set_to("icon")
                element.data.set_active("icon")
            else:
                element.type.set_to("text")
                element.data.set_active("text")
                curr_text += c

            i += 1
            element_type = element.type.enum_name
            while (i < len(message_string) and
                   len(curr_text) < 254 and
                   len(icon_name) < MAX_ICON_NAME_LENGTH):
                c = message_string[i]
                if c == "%":
                    break

                i += 1
                if element_type == "text":
                    curr_text += c
                else:
                    icon_name += c
                    if icon_name in icon_type_map:
                        break

            if element_type == "text":
                # add the delimiter
                element.data.text.data = len(curr_text) + 1
                text_blob += curr_text + "\x00"
            elif icon_name in icon_type_map:
                element.data.icon.data = icon_type_map[icon_name]
            else:
                print("    WARNING: Unknown icon type specified in message '%s'" % name)
                i = element_base + 1
                del elements[-1]
                message.element_count -= 1

    if len(elements) > 8192:
        print("    ERROR: Too many message elements. " +
              "Please simplify your messages.")
        error = True

    if len(text_blob) > 32768:
        print(("    ERROR: String data too large by %s characters. " +
              "Please simplify your messages.") %
              (len(text_blob) - 32768))
        error = True

    # replace the string data
    tagdata.string.data = text_blob

    return error


def compile_unicode_string_list(ustr_tag, string_data):
    return compile_strings(ustr_tag, string_data, True)


def compile_string_list(str_tag, string_data):
    return compile_strings(str_tag, string_data, False)


def compile_strings(tag, string_data, unicode=False):
    if unicode:
        tag_ext = "unicode_string_list"
        tag_cls = "ustr"
        max_str_len = 32768
    else:
        tag_ext = "string_list"
        tag_cls = "str#"
        max_str_len = 4096

    string_data = string_data.replace("\r\n", "\n").\
                  replace("\n\r", "\n").replace("\r", "\n")

    strings = string_data.split("\n###END-STRING###\n")
    if len(strings) > 32767:
        print("    WARNING: Too many strings. Truncating to 32767.")
        del strings[32768: ]

    for i in range(len(strings)):
        string = strings[i]
        if len(strings) > max_str_len:
            print(("    WARNING: String %s is too long." +
                   " Truncating to %s characters") % (i, max_str_len))
            strings[i] = string[max_str_len: ]

    tag_strings = tag.data.tagdata.strings.STEPTREE
    del tag_strings[:]

    for string in strings:
        tag_strings.append()
        tag_string = tag_strings[-1]
        tag_string.data = string



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
from math import log, ceil

__all__ = ("TagIndexManager", "TagDirectoryNode")

TAG_ID_HEADER_STR = "  id  "
INDEXED_HEADER_STR = "| indexed "
FILEPATH_HEADER_STR = "| filepath"


class TagIndexManager:
    _tag_index = ()
    _tag_index_map = ()

    _get_dir_nodes_root = None

    def __init__(self, tag_index, tag_index_map=None):
        self._tag_index = tag_index
        if tag_index_map:
            self._tag_index_map = dict(tag_index_map)
        else:
            self._tag_index_map = {}
            for i in range(len(tag_index)):
                self._tag_index_map[tag_index[i].id & 0xFFff] = i

        # 0xFFff is reserved to mean NULL, so don't map it
        self._tag_index_map.pop(0xFFff, None)

        self._get_dir_nodes_root = TagDirectoryNode(self.tag_index).root_getter

    @property
    def tag_index(self): return self._tag_index
    @property
    def directory_nodes(self): return self._get_dir_nodes_root()

    def get_total_dir_count(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).total_dir_count
    def get_total_file_count(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).total_file_count
    def get_dir_count(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).dir_count
    def get_file_count(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).file_count
    def get_dir_names(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).dir_names
    def get_file_names(self, dir=""):
        return self.directory_nodes.get_dir_node(dir).file_names

    # wrappers around the directory nodes
    def rename_tag(self, tag_path, new_path):
        return self.directory_nodes.rename_tag(tag_path, new_path)

    def rename_tag_by_id(self, tag_id, new_path):
        return self.directory_nodes.rename_tag_by_id(tag_id, new_path)

    def rename_dir(self, curr_dir, new_dir):
        return self.directory_nodes.rename_dir(curr_dir, new_dir)

    def pprint(self, dir="", **kw):
        return self.directory_nodes.pprint(dir, **kw)

    def pprint_files(self, dir="", **kw):
        return self.directory_nodes.pprint_files(dir, **kw)

    def walk(self, top_down=True):
        yield from self.directory_nodes.walk(top_down)

    def translate_tag_id(self, tag_id):
        tag_id &= 0xFFff
        if tag_id in self._tag_index_map:
            return self._tag_index_map[tag_id]
        return tag_id

    def get_tag_index_ref(self, tag_id):
        tag_id = self.translate_tag_id(tag_id)
        if tag_id in range(len(self._tag_index)):
            return self._tag_index[tag_id]


class TagDirectoryNode:
    class SharedRoot:
        __slots__ = ("_val", )
        def __init__(self, val): self._val = val
        def get(self): return self._val
        def set(self, new_val): self._val = new_val

    __slots__ = (
        "_names_to_ids",
        "_ids_to_names",
        "_sub_nodes",
        "_tag_index",
        "_shared_root",
        "remove_empty"
        )
    def __init__(self, tag_index, dir_items=(), shared_root=None):
        self._names_to_ids = {}
        self._ids_to_names = {}
        self._sub_nodes = {}
        self._tag_index = tag_index
        self.remove_empty = True

        if shared_root is None:
            self._shared_root = self.SharedRoot(self)
        else:
            self._shared_root = shared_root

        if self.root is self:
            # this is the root node. add entire tag index as children
            self.recalculate_nodes()
        else:
            for tag_id in dir_items:
                self._add_node(tag_id, dir_items[tag_id])

    def __getitem__(self, item_key):
        if isinstance(item_key, int):
            if item_key in range(len(self.tag_index)):
                return self.tag_index[item_key]
            raise IndexError(
                'Tag_id "%s" is outside the bounds of the tag index.' % item_key)
        elif item_key in self._sub_nodes:
            return self._sub_nodes[item_key]
        elif item_key in self._names_to_ids:
            return self._tag_index[self._names_to_ids[item_key]]
        else:
            item_key = item_key.replace('/', '\\').strip("\\ ")
            return self._get_node(
                [name for name in item_key.lower().split("\\") if name])

    def __str__(self): return self.pprint()

    @property
    def total_file_count(self):
        return self.file_count + sum(node.total_file_count for node in
                                     self._sub_nodes.values())
    @property
    def total_dir_count(self):
        return self.dir_count + sum(node.total_dir_count for node in
                                    self._sub_nodes.values())
    @property
    def file_count(self): return len(self._names_to_ids)
    @property
    def dir_count(self): return len(self._sub_nodes)
    @property
    def file_names(self): return sorted(self._names_to_ids)
    @property
    def dir_names(self): return sorted(self._sub_nodes)

    @property
    def names_to_ids(self): return dict(self._names_to_ids)
    @property
    def ids_to_names(self): return dict(self._ids_to_names)
    @property
    def sub_nodes(self): return dict(self._sub_nodes)
    @property
    def tag_index(self): return self._tag_index
    @property
    def root(self): return self.root_getter()
    @property
    def root_getter(self): return self._shared_root.get
    @property
    def shallow_empty(self): return not(self._names_to_ids or self._sub_nodes)
    @property
    def empty(self):
        if not self.shallow_empty:
            return False

        for node in self._sub_nodes.values():
            if not node.empty:
                return False
        return True

    def get_dir_node(self, dir=""):
        return self._get_node([n for n in dir.replace("/", "\\").
                               lower().strip("\\ ").split("\\") if n])

    def pprint(self, dir="", **kw):
        if dir:
            dir_pieces = [n for n in dir.replace("/", "\\").
                          lower().strip("\\ ").split("\\") if n]
            return self._get_node(dir_pieces).pprint(**kw)

        # these are user-supplied settings
        printout = kw.pop("do_printout", False)
        print_header = kw.get("header", True)
        print_guides = kw.get("guides", True)
        print_files = kw.get("files", True)
        print_tag_ids = kw.get("tag_ids", True) and print_files
        print_indexed = kw.get("indexed", False) and print_files
        dirs_first = kw.get("dirs_first", True)
        extra_dir_spacing = kw.get("extra_dir_spacing", 1)
        depth = kw.get("depth", 0)
        indent = kw.get("indent", 4)

        extra_returns = kw.pop("_extra_returns", False)
        indent_str = kw.get("indent_str", "")
        seen = kw.setdefault("seen", set())

        if depth is None:
            depth = float("inf")

        if id(self) in seen:
            return ""

        seen.add(id(self))
        string = indexed_pad_str = tag_id_pad = ""

        print_files = print_files and self._names_to_ids

        # subtract 1 char for the directory bar column and 1 for the space
        indent = max(indent - (2 if print_guides else 1), 0)

        # increase the indent str for next level
        kw["header"] = False
        kw["_extra_returns"] = int(extra_dir_spacing)
        last_item_indent_str = indent_str + (" " * (indent + 1)) + " "

        if print_tag_ids:
            tag_id_pad = " " * 6

        if print_header:
            if print_tag_ids:
                string += TAG_ID_HEADER_STR
            if print_indexed:
                string += INDEXED_HEADER_STR
            string += FILEPATH_HEADER_STR + "\n"

        if print_indexed:
            indexed_pad_str = " " * len(INDEXED_HEADER_STR)

        # put together tag path strings
        if print_files and not dirs_first:
            string += self.pprint_files(**kw)
        else:
            files_kw = dict(kw)

        dir_indent_str = indent_str
        if print_guides:
            indent_str += "|"
            if not self.file_count or depth > 0:
                dir_indent_str += "|"
            else:
                dir_indent_str += "+"

        kw["indent_str"] = indent_str + (" " * indent) + " "
        kw["depth"] = depth - 1

        dir_indent_str += (("-" if print_guides else " ") * indent) + " "
        # put together sub-directory strings
        item_count = self.dir_count
        for dir_name in sorted(self._sub_nodes):
            item_count -= 1
            if item_count == 0 and not(print_files and dirs_first):
                # helps with readability
                kw.update(_extra_returns=False, indent_str=last_item_indent_str)

            string += "%s%s%s%s\n" % (tag_id_pad, indexed_pad_str,
                                      dir_indent_str, dir_name)
            if depth > 0:
                string += self._sub_nodes[dir_name].pprint(**kw)

        if print_files and dirs_first:
            string += self.pprint_files(**files_kw)

        # add extra spacing after last directory
        for i in range(extra_returns):
            string += tag_id_pad + indexed_pad_str + last_item_indent_str + "\n"

        # either print the string, or return it
        if printout:
            lines = list(string.split("\n"))
            if not lines[-1]:
                lines.pop(-1)
            for line in lines:
                print(line)

            return ""
        else:
            return string

    def pprint_files(self, dir="", **kw):
        if dir:
            dir_pieces = [n for n in dir.replace("/", "\\").
                          lower().strip("\\ ").split("\\") if n]
            return self._get_node(dir_pieces).pprint_files(**kw)

        # these are user-supplied settings
        printout = kw.pop("do_printout", False)
        print_header = kw.get("header", True)
        print_guides = kw.get("guides", True)
        print_tag_ids = kw.get("tag_ids", True)
        print_indexed = kw.get("indexed", False)
        indent = kw.get("indent", 4)

        indent_str = kw.get("indent_str", "")
        string = ""

        # subtract 1 char for the directory bar column and 1 for the space
        indent = max(indent - (2 if print_guides else 1), 0)

        if print_tag_ids:
            tag_id_pad = " " * 6
        else:
            tag_id_pad = ""

        if print_header:
            if print_tag_ids:
                string += TAG_ID_HEADER_STR
            if print_indexed:
                string += INDEXED_HEADER_STR
            string += FILEPATH_HEADER_STR + "\n"

        if print_guides:
            indent_str += "|"

        indent_str += (("-" if print_guides else " ") * indent) + " "
        for name in sorted(self._names_to_ids):
            tag_id = self._names_to_ids[name]
            if print_tag_ids:
                tag_id_str = str(tag_id)
                string += tag_id_str + tag_id_pad[len(tag_id_str): ]

            if print_indexed:
                try:
                    indexed = self.tag_index[tag_id].indexed
                except Exception:
                    indexed = False

                if indexed:
                    string += "   YES" + (" " * (len(INDEXED_HEADER_STR) - 6))
                else:
                    string += " " * len(INDEXED_HEADER_STR)

            string += "%s%s\n" % (indent_str, name)

        # either print the string, or return it
        if printout:
            lines = list(string.split("\n"))
            if not lines[-1]:
                lines.pop(-1)
            for line in lines:
                print(line)

            return ""
        else:
            return string

    def rename_tag(self, curr_path, new_path):
        curr_path = curr_path.replace('/', '\\').strip("\\ ")
        self._rename_tag(
            [name for name in curr_path.lower().split("\\") if name], new_path)

    def rename_tag_by_id(self, tag_id, new_path):
        tag_ref = self.tag_index[tag_id]
        if tag_ref.id == 0xFFffFFff:
            return

        curr_path = tag_ref.path.replace('/', '\\').strip("\\ ")
        curr_path = "%s.%s" % (new_path, tag_ref.class_1.enum_name)
        self._rename_tag(curr_path.lower().split("\\"), new_path)

    def rename_dir(self, curr_dir, new_dir):
        curr_dir = curr_dir.replace('/', '\\').strip("\\ ")
        self._rename_dir(
            [name for name in curr_dir.lower().split("\\") if name], new_dir)

    def walk(self, top_down=True):
        yield from self._walk(top_down, "")

    def recalculate_nodes(self):
        if self.root is not self:
            return self.root.recalculate_nodes()

        self._names_to_ids = {}
        self._ids_to_names = {}
        self._sub_nodes = {}
        tag_id = 0
        for b in self.tag_index:
            if b.id != 0xFFffFFff:
                tag_path = b.path.replace('/', '\\').strip("\\ ").lower()
                self._add_node(tag_id, [
                    name for name in tag_path.lower().split("\\") if name])
            tag_id += 1

    def _apply_names_to_tag_index(self, parent_dir=""):
        # rename tags
        for name in self._names_to_ids:
            tag_path = "\\".join((parent_dir, os.path.splitext(name)[0]))
            self.tag_index[self._names_to_ids[name]].path = tag_path

        # rename directories
        for name, sub_node in self._sub_nodes.items():
            sub_node._apply_names_to_tag_index("\\".join((parent_dir, name)))

    def _get_node(self, node_path_pieces, dir_index=0):
        if dir_index == len(node_path_pieces):
            return self

        name = node_path_pieces[dir_index]
        dir_index += 1
        if dir_index < len(node_path_pieces):
            # continue navigating directories downward
            if name not in self._sub_nodes:
                raise KeyError('Directory "%s" does not exist in "%s"' % (
                    name, "\\".join(node_path_pieces[: dir_index - 1])))

            return self._sub_nodes[name]._get_node(node_path_pieces, dir_index)
        elif name in self._sub_nodes:
            return self._sub_nodes[name]
        elif name in self._names_to_ids:
            return self._tag_index[self._names_to_ids[name]]
        else:
            raise KeyError('Directory/tag name "%s" does not exist in "%s"' % (
                name, "\\".join(node_path_pieces[: dir_index - 1])))

    def _add_dir(self, dir_pieces, dir_node=None, dir_index=0):
        name = None
        if dir_pieces:
            name = dir_pieces[dir_index]
            dir_index += 1
            if name not in self._sub_nodes:
                # plop a new directory in if one doesnt already exist
                self._sub_nodes[name] = TagDirectoryNode(
                    self.tag_index, (), self._shared_root)
            node = self._sub_nodes[name]
        else:
            node = self

        if dir_index < len(dir_pieces):
            # continue navigating directories downward
            node._add_dir(dir_pieces, dir_node, dir_index)
        elif dir_node is not None:
            # update the currently existing directories contents
            node._update(dir_node)
        else:
            # just added an empty dir_node in the previous "if" statement.
            # nothing to do but return
            pass

    def _rename_dir(self, curr_dir_pieces, new_dir, dir_index=0):
        if not curr_dir_pieces:
            # oh boy, we're renaming the root. we'll need to make a
            # new root and add this node to it at the new directory
            new_root = TagDirectoryNode(self.tag_index, (), self._shared_root)
            new_dir = new_dir.replace('/', '\\').strip("\\ ")
            new_dir_pieces = [name for name in new_dir.lower().split("\\") if name]
            try:
                self._shared_root.set(new_root)
                new_root._add_dir(new_dir_pieces, self)
            except Exception:
                self._shared_root.set(self)
                raise
            return

        name = curr_dir_pieces[dir_index]
        dir_index += 1

        if dir_index == len(curr_dir_pieces):
            # rename this directory
            if name not in self._sub_nodes:
                raise KeyError('Directory "%s" does not exist in "%s"' % (
                    name, "\\".join(curr_dir_pieces[: dir_index - 1])))

            new_dir = new_dir.replace('/', '\\').strip("\\ ")
            new_dir_pieces = [name for name in new_dir.lower().split("\\") if name]

            # apply renames to the tags in this directory node
            dir_node = self._sub_nodes.pop(name)
            try:
                dir_node._apply_names_to_tag_index("\\".join(new_dir_pieces))
                self.root._add_dir(new_dir_pieces, dir_node)
            except Exception:
                self._sub_nodes[name] = dir_node
                raise

        elif name in self._sub_nodes:
            # continue navigating directories downward
            self._sub_nodes[name]._rename_dir(
                curr_dir_pieces, new_dir, dir_index)
            if self.root.remove_empty and self._sub_nodes[name].shallow_empty:
                # remove empty directories
                self._sub_nodes.pop(name)
        else:
            raise KeyError('Directory "%s" does not exist in "%s"' % (
                name, "\\".join(curr_dir_pieces[: dir_index - 1])))

    def _add_node(self, tag_id_or_dir_node, tag_path_pieces, dir_index=0):
        name = tag_path_pieces[dir_index]
        dir_index += 1
        if dir_index < len(tag_path_pieces):
            # continue navigating directories downward
            if name not in self._sub_nodes:
                self._sub_nodes[name] = TagDirectoryNode(
                    self.tag_index, (), self._shared_root)

            self._sub_nodes[name]._add_node(
                tag_id_or_dir_node, tag_path_pieces, dir_index)

        elif not isinstance(tag_id_or_dir_node, TagDirectoryNode):
            # add this tag id to this directory
            tag_id = int(tag_id_or_dir_node)
            if self.tag_index[tag_id].id == 0xFFffFFff:
                return

            name_key = "%s.%s" % (name, self.tag_index[tag_id].class_1.enum_name)

            if name_key in self._names_to_ids:
                raise KeyError(
                    'Tag name "%s" already tied to tag_id "%s" in "%s"' %
                    (name_key, self._names_to_ids[name_key],
                     "\\".join(tag_path_pieces[: -1])))

            self._names_to_ids[name_key] = tag_id
            self._ids_to_names[tag_id] = name_key
        elif name in self._sub_nodes:
            # update an existing directory node with this one
            self._sub_nodes[name]._update(tag_id_or_dir_node)
        else:
            # add this directory node to the list of nodes
            self._sub_nodes[name] = tag_id_or_dir_node

    def _rename_tag(self, tag_path_pieces, new_path, dir_index=0):
        name = tag_path_pieces[dir_index]
        dir_index += 1
        if dir_index == len(tag_path_pieces):
            # rename this tag
            if name not in self._names_to_ids:
                raise KeyError('Tag name "%s" does not exist in "%s"' % (
                    name, "\\".join(tag_path_pieces[: dir_index - 1])))

            # pop the name out of names_to_ids and the id out of ids_to_names
            new_path = new_path.replace('/', '\\').strip("\\ ")
            tag_id = self._names_to_ids[name]
            tag_ref = self.tag_index[tag_id]

            old_tag_path = tag_ref.path
            old_name = self._ids_to_names[tag_id]
            new_tag_path_pieces = [
                name for name in new_path.lower().split("\\") if name]
            new_tag_path_pieces[-1] = os.path.splitext(new_tag_path_pieces[-1])[0]

            new_tag_path = "\\".join(new_tag_path_pieces)
            try:
                self._names_to_ids.pop(old_name)
                self._ids_to_names.pop(tag_id)
                self.root._add_node(tag_id, new_tag_path_pieces)
                tag_ref.path = new_tag_path
            except Exception:
                # revert changes and re-raise the exception
                if old_name is not None:
                    self._names_to_ids[old_name] = tag_id
                    self._ids_to_names[tag_id] = old_name
                tag_ref.path = old_tag_path
                raise

        elif name in self._sub_nodes:
            # continue navigating directories downward
            self._sub_nodes[name]._rename_tag(
                tag_path_pieces, new_path, dir_index)
            if self.root.remove_empty and self._sub_nodes[name].shallow_empty:
                # remove empty directories
                self._sub_nodes.pop(name)
        else:
            raise KeyError('Directory "%s" does not exist in "%s"' % (
                name, "\\".join(tag_path_pieces[: dir_index - 1])))

    def _update(self, other_node):
        if other_node is self:
            return

        for tag_id, tag_name in other_node._ids_to_names.items():
            self._add_node(tag_id, (os.path.splitext(tag_name)[0], ))

        for dir_name, sub_node in other_node._sub_nodes.items():
            if dir_name in self._sub_nodes:
                self._sub_nodes[dir_name]._update(sub_node)
            else:
                self._sub_nodes[dir_name] = sub_node

    def _walk(self, top_down, root=""):
        dirs = list(sorted(self._sub_nodes))
        files = list(sorted(self._names_to_ids))

        if top_down:
            yield root, dirs, files
        else:
            for dir_name in dirs:
                dir_path = "\\".join((root, dir_name)).lstrip("\\")
                yield from self._sub_nodes[dir_name]._walk(top_down, dir_path)

        if top_down:
            for dir_name in dirs:
                dir_path = "\\".join((root, dir_name)).lstrip("\\")
                yield from self._sub_nodes[dir_name]._walk(top_down, dir_path)
        else:
            yield root, dirs, files


if __name__ == "__main__":
    '''
    from reclaimer.meta.halo_map import get_tag_index
    from supyr_struct.buffer import get_rawdata

    tag_index = get_tag_index(
        get_rawdata(
            filepath="C:\\Users\\Moses\\Desktop\\halo_test\\maps\\putput.map")
        )

    dir_nodes = TagDirectoryNode(tag_index.tag_index)
    dir_nodes.pprint(do_printout=True, depth=1, extra_dir_spacing=False)
    dir_nodes.pprint("weapons\\assault rifle", do_printout=True, depth=None,
                     files=False, indexed=True)
    dir_nodes.pprint("weapons\\assault rifle", do_printout=True, depth=None,
                     dirs_first=1, indexed=True)
    #dir_nodes.rename_dir("weapons\\assault rifle\\",
    #                     "weapons\\assault rifle\\asdf\\")
    #dir_nodes.rename_dir("weapons\\assault rifle\\asdf\\",
    #                     "weapons\\assault rifle\\", )
    #dir_nodes.rename_dir("", "test\\")
    #dir_nodes.pprint(do_printout=True, depth=1, indexed=0)#'''



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

sotr_input_mappings = (
    {NAME: 'clamp_x', GUI_NAME: 'clamp(x)'},
    {NAME: 'one_minus_clamp_x', GUI_NAME: '1 - clamp(x)'},
    {NAME: 'clamp_x_times_two_minus_one', GUI_NAME: '2*clamp(x) - 1'},
    {NAME: 'one_minus_clamp_x_times_two', GUI_NAME: '1 - 2*clamp(x)'},
    {NAME: 'clamp_x_minus_one_half', GUI_NAME: 'clamp(x) - 1/2'},
    {NAME: 'one_half_minus_clamp_x', GUI_NAME: '1/2 - clamp(x)'},
    {NAME: 'x', GUI_NAME: 'x'},
    {NAME: 'minus_x', GUI_NAME: '-x'},
    )
sotr_output_mappings = (
    'identity',
    {NAME: 'scale_by_one_half', GUI_NAME: 'scale by 1/2'},
    {NAME: 'scale_by_two', GUI_NAME: 'scale by 2'},
    {NAME: 'scale_by_four', GUI_NAME: 'scale by 4'},
    {NAME: 'bias_by_minus_one_half', GUI_NAME: 'bias by -1/2'},
    'expand_normal',
    )


sotr_color_inputs = (
    'zero',
    'one',
    'one_half',
    'negative_one',
    'negative_one_half',

    'map_color_0',
    'map_color_1',
    'map_color_2',
    'map_color_3',
    {NAME: 'vertex_color_0', GUI_NAME: 'vertex color 0 / diffuse light'},
    {NAME: 'vertex_color_1', GUI_NAME: 'vertex color 1 / fade(perpendicular)'},
    'scratch_color_0',
    'scratch_color_1',
    'constant_color_0',
    'constant_color_1',

    'map_alpha_0',
    'map_alpha_1',
    'map_alpha_2',
    'map_alpha_3',
    {NAME: 'vertex_alpha_0', GUI_NAME: 'vertex alpha 0 / diffuse light'},
    {NAME: 'vertex_alpha_1', GUI_NAME: 'vertex alpha 1 / fade(perpendicular)'},
    'scratch_alpha_0',
    'scratch_alpha_1',
    'constant_alpha_0',
    'constant_alpha_1',
    )
sotr_color_outputs = (
    'discard',
    {NAME:'scratch_color_0', GUI_NAME: 'scratch color 0 / final color'},
    'scratch_color_1',
    'vertex_color_0',
    'vertex_color_1',
    'map_color_0',
    'map_color_1',
    'map_color_2',
    'map_color_3',
    )
sotr_color_output_functions = (
    'multiply',
    'dot_product',
    )


sotr_alpha_inputs = (
    'zero',
    'one',
    'one_half',
    'negative_one',
    'negative_one_half',

    'map_blue_0',
    'map_blue_1',
    'map_blue_2',
    'map_blue_3',
    {NAME: 'vertex_blue_0', GUI_NAME: 'vertex blue 0 / diffuse light'},
    {NAME: 'vertex_blue_1', GUI_NAME: 'vertex blue 1 / fade(perpendicular)'},
    'scratch_blue_0',
    'scratch_blue_1',
    'constant_blue_0',
    'constant_blue_1',

    'map_alpha_0',
    'map_alpha_1',
    'map_alpha_2',
    'map_alpha_3',
    {NAME: 'vertex_alpha_0', GUI_NAME: 'vertex alpha 0 / diffuse light'},
    {NAME: 'vertex_alpha_1', GUI_NAME: 'vertex alpha 1 / fade(perpendicular)'},
    'scratch_alpha_0',
    'scratch_alpha_1',
    'constant_alpha_0',
    'constant_alpha_1',
    )
sotr_alpha_outputs = (
    'discard',
    {NAME: 'scratch_alpha_0', GUI_NAME: 'scratch alpha 0 / final alpha'},
    'scratch_alpha_1',
    {NAME: 'vertex_alpha_0', GUI_NAME: 'vertex alpha 0 / fog'},
    'vertex_alpha_1',
    'map_alpha_0',
    'map_alpha_1',
    'map_alpha_2',
    'map_alpha_3',
    )


stage = Struct("stage",
    Bool16("flags" ,
        "color_mux",
        "alpha_mux",
        "a_out_controls_color0_animation",
        ),
    Pad(2),

    SEnum16("color0_source", *function_names,
        COMMENT="If set to 'none', color0 is calculated by blending the\n"
                "two bounds below based on the 'color0 anim function'."),
    SEnum16("color0_anim_function", *animation_functions),
    float_sec("color0_anim_period"),
    QStruct("color0_anim_lower_bound", INCLUDE=argb_float),
    QStruct("color0_anim_upper_bound", INCLUDE=argb_float),
    QStruct("color1", INCLUDE=argb_float),

    Struct('color',
        Struct('input_A',
            SEnum16('input', GUI_NAME='', *sotr_color_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_B',
            SEnum16('input', GUI_NAME='', *sotr_color_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_C',
            SEnum16('input', GUI_NAME='', *sotr_color_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_D',
            SEnum16('input', GUI_NAME='', *sotr_color_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),

        Struct('output_AB',
            SEnum16('output', GUI_NAME='', *sotr_color_outputs),
            SEnum16('function', *sotr_color_output_functions),
            ORIENT='h'
            ),
        Struct('output_CD',
            SEnum16('output', GUI_NAME='', *sotr_color_outputs),
            SEnum16('function', *sotr_color_output_functions),
            ORIENT='h'
            ),
        SEnum16('output_AB_CD_mux_sum', *sotr_color_outputs),
        SEnum16('output_mapping', *sotr_output_mappings)
        ),

    Struct('alpha',
        Struct('input_A',
            SEnum16('input', GUI_NAME='', *sotr_alpha_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_B',
            SEnum16('input', GUI_NAME='', *sotr_alpha_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_C',
            SEnum16('input', GUI_NAME='', *sotr_alpha_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),
        Struct('input_D',
            SEnum16('input', GUI_NAME='', *sotr_alpha_inputs),
            SEnum16('mapped_to', *sotr_input_mappings),
            ORIENT='h'
            ),

        SEnum16('output_AB', *sotr_alpha_outputs),
        SEnum16('output_CD', *sotr_alpha_outputs),
        SEnum16('output_AB_CD_mux_sum', *sotr_alpha_outputs),
        SEnum16('output_mapping', *sotr_output_mappings)
        ),

    SIZE=112
    )

map = Struct("map",
    Bool16("flags" ,
        "unfiltered",
        "u_clamped",
        "v_clamped",
        ),
    Pad(2),
    #shader_transformations
    Float("map_u_scale"),
    Float("map_v_scale"),
    Float("map_u_offset"),
    Float("map_v_offset"),
    float_deg("map_rotation"),  # degrees
    float_zero_to_one("map_bias"),
    dependency("bitmap", "bitm"),

    #shader animations
    Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),

    QStruct("rotation_center", INCLUDE=xy_float),
    SIZE=100,
    )

sotr_attrs = Struct("sotr_attrs",
    #Generic Transparent Shader
    Struct("generic_transparent_shader",
        UInt8("numeric_counter_limit", MIN=0, MAX=255, SIDETIP="[0,255]"),
        Bool8("flags", *trans_shdr_properties),
        SEnum16("first_map_type",             *trans_shdr_first_map_type),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode",      *render_fade_mode),
        SEnum16("framebuffer_fade_source",    *function_outputs),
        Pad(2),
        ),

    #Lens Flare
    float_wu("lens_flare_spacing"),  # world units
    dependency("lens_flare", "lens"),
    reflexive("extra_layers", extra_layers_block, 4,
        DYN_NAME_PATH='.filepath'),
    reflexive("maps", map, 4, DYN_NAME_PATH='.bitmap.filepath'),
    reflexive("stages", stage, 7),
    SIZE=68
    )

sotr_body = Struct("tagdata",
    shdr_attrs,
    sotr_attrs,

    SIZE=108,
    )


def get():
    return sotr_def

sotr_def = TagDef("sotr",
    blam_header("sotr"),
    sotr_body,

    ext=".shader_transparent_generic", endian=">", tag_cls=ShdrTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

snde_body = QStruct("tagdata",
    Pad(4),
    UInt16("priority"),
    Pad(2),
    Float("room_intensity"),
    Float("room_intensity_hf"),
    Float("room_rolloff", MIN=0.0, MAX=10.0),
    float_sec("decay_time", MIN=0.1, MAX=20.0),
    Float("decay_hf_ratio", MIN=0.1, MAX=2.0),
    Float("reflections_intensity"),
    float_sec("reflections_delay", MIN=0.0, MAX=0.3),
    Float("reverb_intensity"),
    float_sec("reverb_delay", MIN=0.0, MAX=0.1),
    Float("diffusion"),
    Float("density"),
    Float("hf_reference", MIN=20.0, MAX=20000.0, SIDETIP="Hz"),
    SIZE=72,
    )

def get():
    return snde_def

snde_def = TagDef("snde",
    blam_header('snde'),
    snde_body,

    ext=".sound_environment", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return vcky_def

virtual_key = Struct("virtual_key",
    SEnum16("keyboard_key",
        {NAME:"one", GUI_NAME:"1"},
        {NAME:"two", GUI_NAME:"2"},
        {NAME:"three", GUI_NAME:"3"},
        {NAME:"four", GUI_NAME:"4"},
        {NAME:"five", GUI_NAME:"5"},
        {NAME:"six", GUI_NAME:"6"},
        {NAME:"seven", GUI_NAME:"7"},
        {NAME:"eight", GUI_NAME:"8"},
        {NAME:"nine", GUI_NAME:"9"},
        {NAME:"zero", GUI_NAME:"0"},
        "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
        "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
        "done",
        "shift",
        "capslock",
        "symbols",
        "backspace",
        "left",
        "right",
        "space",
        ),
    SInt16("lowercase_character"),
    SInt16("shift_character"),
    SInt16("caps_character"),
    SInt16("symbols_character"),
    SInt16("shift_caps_character", GUI_NAME="shift+caps character"),
    SInt16("shift_symbols_character", GUI_NAME="shift+symbols character"),
    SInt16("caps_symbols_character", GUI_NAME="caps+symbols character"),
    dependency("unselected_background_bitmap", "bitm"),
    dependency("selected_background_bitmap", "bitm"),
    dependency("active_background_bitmap", "bitm"),
    dependency("sticky_background_bitmap", "bitm"),
    SIZE=80
    )

vcky_body = Struct("tagdata",
    dependency("display_font", "font"),
    dependency("background_bitmap", "bitm"),
    dependency("special_key_labels_string_list", "ustr"),
    reflexive("virtual_keys", virtual_key, 44,
        DYN_NAME_PATH='.keyboard_key.enum_name'),
    SIZE=60,
    )

vcky_def = TagDef("vcky",
    blam_header('vcky', 2),
    vcky_body,

    ext=".virtual_keyboard", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.coll import CollTag
from supyr_struct.defs.tag_def import TagDef

bsp_node_tooltip = (
    "Refers to a leaf node if negative.\n"
    "Add 0x80000000 to get leaf node index."
    )

modifier = Struct("modifier",
    Pad(52),
    )

body = Struct("body",
    Float("maximum_body_vitality"),
    Float("body_system_shock"),

    Pad(52),
    float_zero_to_one("friendly_damage_resistance"),

    Pad(40),
    dependency("localized_damage_effect", "effe"),

    float_zero_to_one("area_damage_effect_threshold"),
    dependency("area_damage_effect", "effe"),

    Float("body_damaged_threshold"),
    dependency("body_damaged_effect", "effe"),
    dependency("body_depleted_effect", "effe"),
    Float("body_destroyed_threshold"),
    dependency("body_destroyed_effect", "effe"),
    )

shield = Struct("shield",
    Float("maximum_shield_vitality"),

    Pad(2),
    SEnum16("shield_material_type", *materials_list),

    Pad(24),
    SEnum16("shield_failure_function", *fade_functions),

    Pad(2),
    Float("shield_failure_threshold"),
    Float("shield_failing_leak_fraction"),

    Pad(16),
    Float("minimum_stun_damage"),
    float_sec("stun_time"),
    float_sec("recharge_time"),

    Pad(112),
    Float("shield_damaged_threshold"),
    dependency("shield_damaged_effect", "effe"),
    dependency("shield_depleted_effect", "effe"),
    dependency("shield_recharging_effect", "effe"),
    Pad(8),
    Float("shield_recharge_rate", VISIBLE=False),
    )

bsp3d_node = QStruct("bsp3d_node",
    SInt32("plane"),
    SInt32("back_child", TOOLTIP=bsp_node_tooltip),
    SInt32("front_child", TOOLTIP=bsp_node_tooltip),
    SIZE=12
    )

plane = QStruct("plane",
    # i, j, and k form a unit vector where d specifies
    # the location of a point a "distance" along it
    Float("i"), Float("j"), Float("k"), Float("d"),
    SIZE=16, ORIENT='h'
    )

leaf = Struct("leaf",
    Bool16("flags",
        "contains_double_sided_surfaces"
        ),
    SInt16("bsp2d_reference_count"),
    SInt32("first_bsp2d_reference"),
    SIZE=8
    )

bsp2d_reference = QStruct("bsp2d_reference",
    SInt32("plane"),
    SInt32("bsp2d_node", TOOLTIP=bsp_node_tooltip),
    SIZE=8
    )

bsp2d_node = QStruct("bsp2d_node",
    Float("plane_i"),
    Float("plane_j"),
    Float("plane_d"),
    SInt32("left_child", TOOLTIP=bsp_node_tooltip),
    SInt32("right_child", TOOLTIP=bsp_node_tooltip),
    SIZE=20
    )

surface = Struct("surface",
    SInt32("plane"),
    SInt32("first_edge"),
    Bool8("flags",
        "two_sided",
        "invisible",
        "climbable",
        "breakable",
        ),
    SInt8("breakable_surface"),
    SInt16("material"),
    SIZE=12
    )

edge = QStruct("edge",
    SInt32("start_vertex"),
    SInt32("end_vertex"),
    SInt32("forward_edge"),
    SInt32("reverse_edge"),
    SInt32("left_surface"),
    SInt32("right_surface"),
    SIZE=24
    )

vertex = QStruct("vertex",
    Float("x"),
    Float("y"),
    Float("z"),
    SInt32("first_edge"),
    SIZE=16
    )

permutation_bsp = Struct("permutation_bsp",
    reflexive("bsp3d_nodes", bsp3d_node, 131072),
    reflexive("planes", plane, 65535),
    reflexive("leaves", leaf, 65535),
    reflexive("bsp2d_references", bsp2d_reference, 131072),
    reflexive("bsp2d_nodes", bsp2d_node, 65535),
    reflexive("surfaces", surface, 131072),
    reflexive("edges", edge, 262144),
    reflexive("vertices", vertex, 131072),
    SIZE=96
    )

node = Struct("node",
    ascii_str32("name"),
    dyn_senum16("region",
        DYN_NAME_PATH=".....regions.regions_array[DYN_I].name"),
    dyn_senum16("parent_node",
        DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16("next_sibling_node",
        DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16("first_child_node",
        DYN_NAME_PATH="..[DYN_I].name"),

    Pad(8),
    FlSInt16("unknown0", VISIBLE=False),
    FlSInt16("unknown1", VISIBLE=False),
    reflexive("bsps", permutation_bsp, 32),
    SIZE=64
    )

pathfinding_sphere = Struct("pathfinding_sphere",
    dyn_senum16("node",
        DYN_NAME_PATH=".....nodes.nodes_array[DYN_I].name"),

    Pad(14),
    QStruct("center", INCLUDE=xyz_float),
    Float("radius"),
    SIZE=32
    )

permutation = Struct("permutation",
    ascii_str32("name"),
    SIZE=32
    )

region = Struct("region",
    ascii_str32("name"),
    Bool32("flags",
        "lives_until_object_dies",
        "forces_object_to_die",
        "dies_when_object_dies",
        "dies_when_object_is_damaged",
        "disappears_when_shield_is_off",
        "inhibits_melee_attack",
        "inhibits_walking",
        "forces_drop_weapon",
        "causes_head_maimed_scream",
        ),
    Pad(4),
    Float("damage_threshold"),

    Pad(12),
    dependency("destroyed_effect", "effe"),
    reflexive("permutations", permutation, 32, DYN_NAME_PATH='.name'),
    SIZE=84
    )

material = Struct("material",
    ascii_str32("name"),
    Bool32("flags",
        "head"
        ),
    SEnum16("material_type", *materials_list),
    Pad(2),
    Float("shield_leak_percentage"),
    Float("shield_damage_multiplier"),

    Pad(12),
    Float("body_damage_multiplier"),
    SIZE=72
    )

coll_body = Struct("tagdata",
    Bool32("flags",
        "takes_shield_damage_for_children",
        "takes_body_damage_for_children",
        "always_shields_friendly_damage",
        "passes_area_damage_to_children",
        "parent_never_takes_body_damage_for_us",
        "only_damaged_by_explosives",
        "only_damaged_while_occupied",
        ),
    dyn_senum16("indirect_damage_material",
        DYN_NAME_PATH=".materials.materials_array[DYN_I].name"),
    Pad(2),

    body,
    shield,

    Pad(112),
    reflexive("materials", material, 32, DYN_NAME_PATH='.name'),
    reflexive("regions", region, 8, DYN_NAME_PATH='.name'),
    reflexive("modifiers", modifier, 0, VISIBLE=False),

    Pad(16),
    Struct("pathfinding_box",
        QStruct("x", INCLUDE=from_to),
        QStruct("y", INCLUDE=from_to),
        QStruct("z", INCLUDE=from_to),
        ),

    reflexive("pathfinding_spheres", pathfinding_sphere, 32),
    reflexive("nodes", node, 64, DYN_NAME_PATH='.name'),

    SIZE=664,
    )


fast_permutation_bsp = Struct("permutation_bsp",
    raw_reflexive("bsp3d_nodes", bsp3d_node, 131072),
    raw_reflexive("planes", plane, 65535),
    raw_reflexive("leaves", leaf, 65535),
    raw_reflexive("bsp2d_references", bsp2d_reference, 131072),
    raw_reflexive("bsp2d_nodes", bsp2d_node, 65535),
    raw_reflexive("surfaces", surface, 131072),
    raw_reflexive("edges", edge, 262144),
    raw_reflexive("vertices", vertex, 131072),
    SIZE=96
    )

fast_node = Struct("node",
    ascii_str32("name"),
    dyn_senum16("region",
        DYN_NAME_PATH=".....regions.regions_array[DYN_I].name"),
    dyn_senum16("parent_node",
        DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16("next_sibling_node",
        DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16("first_child_node",
        DYN_NAME_PATH="..[DYN_I].name"),

    Pad(8),
    FlSInt16("unknown0", VISIBLE=False),
    FlSInt16("unknown1", VISIBLE=False),
    reflexive("bsps", fast_permutation_bsp, 32),
    SIZE=64
    )

fast_coll_body = dict(coll_body)
fast_coll_body[12] = reflexive("nodes", fast_node, 64, DYN_NAME_PATH='.name')


def get():
    return coll_def

coll_def = TagDef("coll",
    blam_header("coll", 10),
    coll_body,

    ext=".model_collision_geometry", endian=">", tag_cls=CollTag,
    )

fast_coll_def = TagDef("coll",
    blam_header("coll", 10),
    fast_coll_body,

    ext=".model_collision_geometry", endian=">", tag_cls=CollTag,
    )



# Module taken from https://github.com/MosesofEgypt/snippets
#
# ISC License (ISC)
#
# Copyright 2017 Devin Bobadilla
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

from math import sqrt

DEFAULT_TEX = 0
MAX_STRIP_LEN = 2**32-4


class StripTri(list):
    __slots__ = (
        "siblings", "sibling_edges", "added",
        "get_sibling_index", "get_edge_index"
        )

    def __init__(self, v0, v1, v2):
        list.__init__(self, (v0, v1, v2))
        self.added = False
        self.siblings = [None, None, None]
        self.sibling_edges = [(v1, v0, 0),
                              (v2, v1, 0),
                              (v0, v2, 0)]

    def __setattr__(self, attr_name, new_val):
        if attr_name == "sibling_edges":
            self.get_edge_index = new_val.index
        elif attr_name == "siblings":
            self.get_sibling_index = new_val.index
        object.__setattr__(self, attr_name, new_val)

    def __add__(self, new_val): raise NotImplementedError()
    def __iadd__(self, new_val): raise NotImplementedError()
    def append(self): raise NotImplementedError()
    def extend(self): raise NotImplementedError()


class Stripifier():
    ''''''
    # the max length a strip can be
    max_strip_len = MAX_STRIP_LEN

    # whether or not the strips have all been linked together
    linked = False

    _winding = False

    # don't edit any of these directly if you know what's good for you
    vert_map = ()
    vert_data = ()
    all_tris_by_edges = ()
    all_face_dirs = ()
    all_strips = ()
    tri_counts = ()

    def __init__(self, all_tris=None, winding=False, *args, **kwargs):
        '''class initialization'''
        self.load_mesh(all_tris, winding)

    def translate_strip(self, vert_indices):
        '''
        Given a triangle strip that this Stripifier instance
        calculated, this method will return a de-indexed triangle strip.
        The elements in the strip will be the vert data the triangles
        contained when the current mesh was loaded.'''
        vert_data = self.vert_data
        return [vert_data[i][0] for i in vert_indices]

    def get_strip(self, strip_i=0, tex_index=DEFAULT_TEX):
        if tex_index not in self.all_strips:
            raise ValueError("No texture index '%s'" % tex_index)
        elif strip_i not in range(len(self.all_strips[tex_index])):
            raise ValueError("Strip index '%s' not in range(%s)" %
                             (strip_i, len(self.all_strips[tex_index])))

        return self.all_strips[tex_index][strip_i]

    def get_strip_count(self, tex_index=None):
        '''
        Returns the number of strips for the specified texture.
        If unspecified, returns the number of all loaded strips.'''
        indices = (tex_index, ) if tex_index is not None else self.all_strips
        ct = 0
        for i in indices:
            ct += len(self.all_strips[i])
        return ct

    def calc_strip(self, tri, neighbor_i=0, set_added=True):
        '''
        Given a starting triangle and the edge to start navigating,
        this function will return a list of the verts that make up the
        longest strip it can find and whether the strip faces backward.

        If set_added is True, triangles found will be flagged as
        added to a strip. Otherwise they will be de-flagged.'''
        vert_data = self.vert_data
        # each triangle has 3 edges.
        # this is the index of the edge the
        # next triangle will be connected to
        neighbor_i = neighbor_i%3

        strip_dir = self._winding

        seen = set()
        set_added = bool(set_added)

        '''navigate the strip in reverse to find the best place to start'''
        while True:
            # set the last triangle as this one
            last_tri = tri
            tri = tri.siblings[neighbor_i]

            # exit if the strip has ended
            if tri is None or tri.added or id(tri) in seen:
                tri = last_tri
                break

            # get which index the last tri was in the new tri so we can
            # orient outselves and figure out which edge to travel next
            neighbor_i = (tri.get_sibling_index(last_tri) + 1 + strip_dir) % 3

            # reverse the direction of travel
            # and set the triangle as seen
            strip_dir = not strip_dir
            seen.add(id(last_tri))

        # reset the seen set
        seen = set()

        # make a strip starting with the first 2 verts to the triangle
        strip = [tri[neighbor_i], tri[(neighbor_i + 1) % 3]]
        if strip_dir == self._winding:
            strip = strip[::-1]

        '''loop over triangles until the length is maxed or
        we reach a triangle without a neighbor on that edge'''
        curr_dir = strip_dir
        while not(tri.added or id(tri) in seen or len(strip) > self.max_strip_len):
            # get the index of the vert that will be added to the strip
            v_i = tri[(neighbor_i + 2) % 3]

            # add the vert to the strip
            strip.append(v_i)

            # set the last triangle as this one
            last_tri = tri

            # Get the next triangle.
            tri = tri.siblings[(neighbor_i + 1 + curr_dir) % 3]

            # reverse the direction of travel, set the last triangle
            # as added and seen, and increment the strip length
            last_tri.added = set_added
            curr_dir = not curr_dir
            seen.add(id(last_tri))

            # exit if the strip has ended
            if tri is None: break

            # get which index the last tri was in the new tri so we can
            # orient outselves and figure out which edge to travel next
            neighbor_i = tri.get_sibling_index(last_tri)

        return strip, strip_dir

    def link_strips(self):
        '''
        Links the strips that are currently loaded together into one
        large strip. This will introduce degenerate triangles into the
        mesh to link strips.'''
        if self.linked:
            return

        all_strips = self.all_strips
        all_face_dirs = self.all_face_dirs

        max_len = self.max_strip_len
        winding = self._winding

        for tex_index in all_strips:
            strips = all_strips[tex_index]
            face_dirs = all_face_dirs[tex_index]

            # if there are no strips for this mesh, skip it
            if not len(strips):
                continue

            even_fore_strips = []
            even_back_strips = []
            odd_strips = []
            strip_ct = len(strips)

            for i in range(strip_ct):
                strip = strips[i]
                face_dir = face_dirs[i]
                is_even = not (len(strip) & 1)
                if face_dir:
                    if is_even:
                        if winding and len(strip) == 4:
                            even_fore_strips.append(
                                [strip[0], strip[2], strip[1], strip[3]])
                        else:
                            even_back_strips.append(strip)
                    elif winding:
                        odd_strips.append(strip)
                    else:
                        odd_strips.append(list(reversed(strip)))
                elif is_even:
                    if winding and len(strip) == 4:
                        even_back_strips.append(
                            [strip[0], strip[2], strip[1], strip[3]])
                    else:
                        even_fore_strips.append(strip)
                elif winding:
                    odd_strips.append(list(reversed(strip)))
                else:
                    odd_strips.append(strip)

            fully_sorted_strips = [None] * strip_ct
            fully_sorted_face_dirs = [None] * strip_ct
            facing = True
            all_even_strips = (even_back_strips, even_fore_strips)
            if not winding:
                facing = False
                all_even_strips = (even_fore_strips, even_back_strips)

            i = 0
            for even_strips in all_even_strips:
                for s in even_strips:
                    fully_sorted_strips[i] = s
                    fully_sorted_face_dirs[i] = facing
                    i += 1
                facing = not facing

            for j in range(len(odd_strips)):
                fully_sorted_strips[i] = odd_strips[j]
                fully_sorted_face_dirs[i] = winding
                i += 1

            # make lists to hold the new strips, degens, and dirs
            new_strips = []
            new_face_dirs = []

            # get the first set of strips to link together
            strip_i = 1
            strip_ct = len(fully_sorted_strips)

            # get the first strip to link together
            strip0 = fully_sorted_strips[0]
            strip0_dir = fully_sorted_face_dirs[0]

            '''keep linking strips together till none are left'''
            while strip_i < strip_ct or not new_strips:
                strip1 = strip1_dir = None
                if strip0_dir != winding:
                    strip0.insert(0, strip0[0])
                    strip0_dir = not strip0_dir

                # link strips together until their length is maxed
                while strip_i < strip_ct:
                    strip1 = fully_sorted_strips[strip_i]
                    strip1_dir = fully_sorted_face_dirs[strip_i]

                    len0 = len(strip0)
                    len1 = len(strip1)
                    end_dir0 = (strip0_dir == (len0 % 2))
                    add_degen = (end_dir0 == strip1_dir)

                    strip_i += 1
                    # if the strip being added is empty, skip it
                    if len0 + len1 + add_degen + 2 > max_len:
                        # total length will be over max. dont try to combine
                        break
                    elif len1 == 0:
                        continue

                    # if a degen needs to be added, repeat the last vert
                    if add_degen:
                        strip0.append(strip0[-1])

                    if strip0[-1] != strip1[0]:
                        # create the extra degenerate triangles
                        strip0.extend((strip0[-1], strip1[0]))

                    # merge the strips
                    strip0.extend(strip1)

                new_face_dirs.append(strip0_dir)
                new_strips.append(strip0)

                # restart the cycle on the next strip
                strip0_dir = strip1_dir
                strip0 = strip1

            all_strips[tex_index]    = new_strips
            all_face_dirs[tex_index] = new_face_dirs

        self.linked = True

    def load_mesh(self, all_tris=None, winding=False):
        '''Loads a list of triangles into the stripifier.'''
        # Maps each unique vert/uv/norm/color combo to
        # the index it is stored in in vert_data
        self.vert_map = vert_map = {}
        # Stores each unique combination of vert/uv/norm/color number
        self.vert_data = vert_data = []

        '''Stores triangles indexed by their tex_index and then their edges.
        The triangle edges they are indexed under are in reverse direction.
        This is because all connected neighboring triangles will
        share the same edge, but in the opposite direction.'''
        self.all_tris_by_edges = {}

        '''Stores lists of the direction of each tex_indexs triangle strips.
        False == strip is facing properly. True == strip is facing inverted.'''
        self.all_face_dirs = {}

        '''Stores the triangle strip lists indexed by their tex_index'''
        self.all_strips = {}

        '''Stores tri counts for each subobject separated by tex_index'''
        self.tri_counts = {}

        # whether or not the strips have been linked together
        self.linked = False

        self._winding = winding

        if all_tris is None:
            return
        elif isinstance(all_tris, (list, tuple)):
            all_tris = {DEFAULT_TEX:all_tris}
        elif not isinstance(all_tris, dict):
            raise TypeError("'all_tris' argument must be either a list"
                            ", tuple, or dict, not %s" % type(all_tris))

        # loop over all meshes by texture
        for tex_index in all_tris:
            tris = all_tris[tex_index]

            self.all_tris_by_edges[tex_index] = t_by_e = {}
            self.all_face_dirs[tex_index] = []
            self.all_strips[tex_index] = []
            self.tri_counts[tex_index] = len(tris)
            if not tris:
                continue

            make_tuple = (hasattr(tris[0], "__getitem__") and
                          hasattr(tris[0][0], "__iter__") and
                          not isinstance(tris[0][0], tuple))

            for src_tri in tris:
                if make_tuple:
                    v0 = (tuple(src_tri[0]), tex_index)
                    v1 = (tuple(src_tri[1]), tex_index)
                    v2 = (tuple(src_tri[2]), tex_index)
                else:
                    v0 = (src_tri[0], tex_index)
                    v1 = (src_tri[1], tex_index)
                    v2 = (src_tri[2], tex_index)

                # connected faces in a triangle strip must share vert,
                # coordinates, uv coordinates, and normals. This is
                # because the verts are reused for neighboring faces.
                # Need to split strips up by texture coordinates as well.

                v0_i = vert_map.get(v0)
                # get if this first vert doesnt already exist
                if v0_i is None:
                    vert_map[v0] = v0_i = len(vert_data)
                    vert_data.append(v0)

                v1_i = vert_map.get(v1)
                # get if this second vert doesnt already exist
                if v1_i is None:
                    vert_map[v1] = v1_i = len(vert_data)
                    vert_data.append(v1)

                v2_i = vert_map.get(v2)
                # get if this third vert doesnt already exist
                if v2_i is None:
                    vert_map[v2] = v2_i = len(vert_data)
                    vert_data.append(v2)

                tri = StripTri(v0_i, v1_i, v2_i)
                edges = ((v0_i, v1_i, 0),
                         (v1_i, v2_i, 0),
                         (v2_i, v0_i, 0))

                # loop over all 3 edges
                for i in (0, 1, 2):
                    v0_i, v1_i, update_edge_num = edge = edges[i]
                    rev_edge = (v1_i, v0_i, update_edge_num)

                    if edge in t_by_e:
                        # this takes care of update edges, which
                        # is when more than two tris share an edge.
                        while edge in t_by_e:
                            update_edge_num += 1
                            edge = (v0_i, v1_i, update_edge_num)

                        rev_edge = (v1_i, v0_i, update_edge_num)
                        tri.sibling_edges[i] = rev_edge

                    # get the triangle that shares this edge
                    conn_tri = t_by_e.get(rev_edge)

                    if conn_tri:
                        # Some triangle shares this edge, so add it
                        # to this triangle as one of its siblings and
                        # add this triangle to it as one of its siblings
                        tri.siblings[i] = conn_tri
                        conn_tri.siblings[conn_tri.get_edge_index(edge)] = tri

                    # neighbor edges are travelled in reverse.
                    t_by_e[edge] = tri

    def make_strips(self):
        '''Creates triangle strips out of the loaded mesh.'''
        all_tris_by_edges = self.all_tris_by_edges

        '''loop over all meshes by texture'''
        for tex_index in all_tris_by_edges:
            tris = all_tris_by_edges[tex_index]
            self.all_face_dirs[tex_index] = face_dirs = []
            self.all_strips[tex_index] = strips = []

            tri_count = self.tri_counts[tex_index]
            edges = list(tris.keys())

            tris_added = 0
            e_i = 0

            '''create triangle strips for this mesh'''
            while tris_added < tri_count:
                # get the first triangle in the strip
                tri_0 = tris[edges[e_i]]
                e_i += 1

                # if the triangle has already been added
                if tri_0.added:
                    continue

                # calculate the 3 different possible strips
                s0, _ = self.calc_strip(tri_0, 0, False)
                s1, _ = self.calc_strip(tri_0, 1, False)
                s2, _ = self.calc_strip(tri_0, 2, False)

                lens = (len(s0), len(s1), len(s2))

                # use only the largest strip
                # Need to re-run the function so it can also
                # flag the triangles in the strip as added.
                strip_data = self.calc_strip(tri_0, lens.index(max(lens)), True)
                strips.append(strip_data[0])
                face_dirs.append(strip_data[1])

                tris_added += len(strips[-1]) - 2



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import log

from reclaimer.hek.defs.objs.tag import HekTag
from reclaimer.util.matrices import Matrix


class PhysTag(HekTag):

    def calc_masses(self):
        data = self.data.tagdata
        mass_points = data.mass_points.STEPTREE
        mass_center = [0,0,0]

        if not len(mass_points):
            data.center_of_mass[:] = mass_center
            return

        total_mass = data.mass
        total_rel_mass = sum([mp.relative_mass for mp in mass_points])

        for mp in mass_points:
            rel_mass = mp.relative_mass
            position = mp.position

            mp.mass = total_mass*(rel_mass/total_rel_mass)
            mass_center[0] += position[0]*mp.mass
            mass_center[1] += position[1]*mp.mass
            mass_center[2] += position[2]*mp.mass

        mass_center[0] /= total_mass
        mass_center[1] /= total_mass
        mass_center[2] /= total_mass

        data.center_of_mass[:] = mass_center

    def calc_densities(self):
        data = self.data.tagdata
        mass_points = data.mass_points.STEPTREE

        if not len(mass_points):
            return

        total_mass = data.mass
        total_density = data.density

        densities = [mp.relative_density for mp in
                     mass_points if mp.relative_density]
        masses = [mp.relative_mass for mp in
                  mass_points if mp.relative_mass]
        if not densities or not masses:
            for mp in mass_points:
                mp.density = 0
            return

        total_rel_mass = sum(mp.relative_mass for mp in mass_points)
        average_rel_mass = total_rel_mass / len(mass_points)
        #                    total_density      total_density
        # density_scale = ------------------ = ----------------
        #                  avg(mass/density)    avg( m^3   kg )
        #                                          ( --- * -- )
        #                                          ( kg    1  )
        den_scale = total_density / len(mass_points)
        den_scale *= sum(mp.relative_mass / mp.relative_density
                         for mp in mass_points if mp.relative_density)
        mass_scale = 1 / average_rel_mass

        for mp in mass_points:
            mp.density = den_scale * mass_scale * mp.relative_density

    def calc_intertia_matrices(self):
        data = self.data.tagdata
        scale = data.moment_scale
        matrices = data.inertia_matrices.STEPTREE
        com = data.center_of_mass
        mass_points = data.mass_points.STEPTREE

        # make sure the matrix array is only 2 long
        matrices.extend(2 - len(matrices))
        del matrices[2:]

        reg = matrices.regular
        inv = matrices.inverse

        reg_yy_zz, reg_zz_xx, reg_xx_yy = reg.yy_zz, reg.zz_xx, reg.xx_yy
        xx = yy = zz = float('1e-30')  # prevent division by 0
        neg_zx = neg_xy = neg_yz = 0

        # calculate the moments for each mass point and add them up
        for mp in mass_points:
            pos = mp.position

            dist_xx = (com[1] - pos[1])**2 + (com[2] - pos[2])**2
            dist_yy = (com[0] - pos[0])**2 + (com[2] - pos[2])**2
            dist_zz = (com[0] - pos[0])**2 + (com[1] - pos[1])**2

            dist_zx = (com[0] - pos[0])*(com[2] - pos[2])
            dist_xy = (com[0] - pos[0])*(com[1] - pos[1])
            dist_yz = (com[1] - pos[1])*(com[2] - pos[2])

            if mp.radius > 0:
                radius_term = 4*pow(10, (2*log(mp.radius, 10) - 1))
            else:
                radius_term = 0

            xx += (dist_xx + radius_term) * mp.mass
            yy += (dist_yy + radius_term) * mp.mass
            zz += (dist_zz + radius_term) * mp.mass
            neg_zx -= dist_zx * mp.mass
            neg_xy -= dist_xy * mp.mass
            neg_yz -= dist_yz * mp.mass

        xx, yy, zz = xx*scale, yy*scale, zz*scale
        neg_zx, neg_xy, neg_yz = neg_zx*scale, neg_xy*scale, neg_yz*scale

        # place the calculated values into the matrix
        reg_yy_zz[:] = xx, neg_xy, neg_zx
        reg_zz_xx[:] = neg_xy, yy, neg_yz
        reg_xx_yy[:] = neg_zx, neg_yz, zz

        # calculate the inverse inertia matrix
        regular = Matrix((reg_yy_zz, reg_zz_xx, reg_xx_yy))
        try:
            inverse = regular.inverse
        except ZeroDivisionError:
            inverse = Matrix((1, 0, 0), (0, 1, 0), (0, 0, 1))
            print("Could not calculate inertia matrix inverse.")

        # place the inverse matrix into the tag
        inv.yy_zz[:] = inverse[0][:]
        inv.zz_xx[:] = inverse[1][:]
        inv.xx_yy[:] = inverse[2][:]

        # copy the xx, yy, and zz moments form the matrix into the tag body
        data.xx_moment = reg_yy_zz[0]
        data.yy_moment = reg_zz_xx[1]
        data.zz_moment = reg_xx_yy[2]

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)
        self.calc_masses()
        self.calc_densities()
        self.calc_intertia_matrices()



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get():
    return boom_def

boom_def = TagDef("boom",
    blam_header('boom'),
    QStruct('tagdata',
        #this is just a guess. This could just as easily
        #be 4 bytes of padding. effing useless tag type
        Float('radius')
        ),

    ext=".spheroid", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sin, cos

from reclaimer.hek.defs.objs.obje import ObjeTag

class BipdTag(ObjeTag):

    def calc_internal_data(self):
        ObjeTag.calc_internal_data(self)
        bipd_attrs = self.data.tagdata.bipd_attrs
        movement = bipd_attrs.movement
        physics = bipd_attrs.physics

        physics.cosine_stationary_turning_threshold = cos(bipd_attrs.stationary_turning_threshold)

        physics.cosine_maximum_slope_angle = cos(movement.maximum_slope_angle)
        physics.neg_sine_downhill_falloff_angle = -sin(movement.downhill_falloff_angle)
        physics.neg_sine_downhill_cutoff_angle  = -sin(movement.downhill_cutoff_angle)
        physics.sine_uphill_falloff_angle = sin(movement.uphill_falloff_angle)
        physics.sine_uphill_cutoff_angle  = sin(movement.uphill_cutoff_angle)

        physics.crouch_camera_velocity = 0
        if physics.crouch_camera_velocity:
            physics.crouch_camera_velocity /= bipd_attrs.camera_collision_and_autoaim.crouch_transition_time

        physics.crouch_camera_velocity /= 30



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


skya_atmosphere_propertie = Struct("atmosphere_propertie",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_string_id("name"),
    Float("light_source_y"),
    Float("light_source_x"),
    color_rgb_float("fog_color"),
    Float("brightness"),
    Float("fog_gradient_threshold"),
    Float("light_intensity"),
    Float("sky_invisiblility_through_fog"),
    Float("unknown_2", VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    Float("light_source_spread"),
    BytesRaw("unknown_4", SIZE=4, VISIBLE=False),
    Float("fog_intensity"),
    Float("unknown_5", VISIBLE=False),
    Float("tint_cyan"),
    Float("tint_magenta"),
    Float("tint_yellow"),
    Float("fog_intensity_cyan"),
    Float("fog_intensity_magenta"),
    Float("fog_intensity_yellow"),
    color_rgb_float("background_color"),
    Float("tint_red"),
    Float("tint2_green"),
    Float("tint2_blue"),
    Float("fog_intensity2"),
    Float("start_distance"),
    Float("end_distance"),
    QStruct("fog_velocity", INCLUDE=xyz_float),
    h3_dependency("weather_effect"),
    BytesRaw("unknown_6", SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=164
    )


skya_underwater = Struct("underwater",
    h3_string_id("name"),
    color_argb_float("color"),
    ENDIAN=">", SIZE=20
    )


skya_body = Struct("tagdata",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_dependency("fog_bitmap"),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=7, VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    h3_reflexive("atmosphere_properties", skya_atmosphere_propertie),
    h3_reflexive("underwater", skya_underwater),
    ENDIAN=">", SIZE=76
    )


def get():
    return skya_def

skya_def = TagDef("skya",
    h3_blam_header('skya'),
    skya_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["skya"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import pack, unpack
from types import MethodType

from reclaimer.meta.halo1_map_fast_functions import iter_reflexive_offs,\
     move_rawdata_ref


def rawdata_ref_move_antr(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    tag_offset = index_array[tag_id].meta_offset
    for moff in iter_reflexive_offs(map_data, tag_offset + 116 - magic, 180):
        # frame info
        move_rawdata_ref(map_data, moff +  72, magic, engine, diffs_by_offsets)
        # default data
        move_rawdata_ref(map_data, moff + 140, magic, engine, diffs_by_offsets)
        # frame data
        move_rawdata_ref(map_data, moff + 160, magic, engine, diffs_by_offsets)


def rawdata_ref_move_bitm(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    packer   = MethodType(pack, "<L")
    unpacker = MethodType(unpack, "<H8xL")
    tag_offset = index_array[tag_id].meta_offset
    for moff in iter_reflexive_offs(map_data, tag_offset + 96 - magic, 48):
        ptr_off = moff - magic

        map_data.seek(ptr_off + 14)
        flags, raw_ptr = unpacker(map_data.read(14))
        if flags & (1<<8):
            # data in resource map
            continue

        ptr_diff = 0
        for off, diff in diffs_by_offsets.items():
            if off <= raw_ptr:
                ptr_diff = diff

        if not ptr_diff:
            continue

        # fix bitmap pointers
        map_data.seek(ptr_off + 24)
        map_data.write(packer(raw_ptr + ptr_diff))


def rawdata_ref_move_devc(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    tag_offset = index_array[tag_id].meta_offset
    # device id
    move_rawdata_ref(map_data, tag_offset +  4, magic, engine, diffs_by_offsets)
    # profile
    move_rawdata_ref(map_data, tag_offset + 24, magic, engine, diffs_by_offsets)


def rawdata_ref_move_font(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    # pixels
    move_rawdata_ref(map_data, index_array[tag_id].meta_offset + 136,
                     magic, engine, diffs_by_offsets)


def rawdata_ref_move_metr(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    # meter data
    move_rawdata_ref(map_data, index_array[tag_id].meta_offset + 152,
                     magic, engine, diffs_by_offsets)


def rawdata_ref_move_scnr(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    tag_offset = index_array[tag_id].meta_offset
    # scenario editor data
    move_rawdata_ref(map_data, tag_offset +  260, magic, engine, diffs_by_offsets)

    # comments
    for moff in iter_reflexive_offs(map_data, tag_offset + 280 - magic, 64):
        move_rawdata_ref(map_data, moff + 44, magic, engine, diffs_by_offsets)

    # recorded animations
    for moff in iter_reflexive_offs(map_data, tag_offset + 876 - magic, 64):
        move_rawdata_ref(map_data, moff + 44, magic, engine, diffs_by_offsets)

    # script syntax data
    move_rawdata_ref(map_data, tag_offset + 1140, magic, engine, diffs_by_offsets)
    # script string data
    move_rawdata_ref(map_data, tag_offset + 1160, magic, engine, diffs_by_offsets)

    # source files
    for moff in iter_reflexive_offs(map_data, tag_offset + 1216 - magic, 52):
        move_rawdata_ref(map_data, moff + 32, magic, engine, diffs_by_offsets)


def rawdata_ref_move_snd_(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    tag_offset = index_array[tag_id].meta_offset
    for moff in iter_reflexive_offs(map_data, tag_offset + 152 - magic, 72):
        for moff2 in iter_reflexive_offs(map_data, moff + 60 - magic, 124):
            # samples
            move_rawdata_ref(map_data, moff2 +  64, magic, engine, diffs_by_offsets)
            # mouth data
            move_rawdata_ref(map_data, moff2 +  84, magic, engine, diffs_by_offsets)
            # subtitle data
            move_rawdata_ref(map_data, moff2 + 104, magic, engine, diffs_by_offsets)


def rawdata_ref_move_shpp(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    # shader code binary
    move_rawdata_ref(map_data, index_array[tag_id].meta_offset + 24,
                     magic, engine, diffs_by_offsets)


def rawdata_ref_move_sidy(tag_id, index_array, map_data, magic, engine,
                          diffs_by_offsets):
    # documentation
    move_rawdata_ref(map_data, index_array[tag_id].meta_offset,
                     magic, engine, diffs_by_offsets)


rawdata_ref_move_functions = {
    "antr": rawdata_ref_move_antr,
    "bitm": rawdata_ref_move_bitm,
    "devc": rawdata_ref_move_devc,
    "font": rawdata_ref_move_font,
    "metr": rawdata_ref_move_metr,
    "scnr": rawdata_ref_move_scnr,
    "snd!": rawdata_ref_move_snd_,

    # open sauce
    "magy": rawdata_ref_move_antr,
    "shpg": rawdata_ref_move_shpp,
    "shpp": rawdata_ref_move_shpp,
    "sidy": rawdata_ref_move_sidy,
    }



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import traceback

from math import pi

from reclaimer.enums import unit_animation_names, unit_weapon_animation_names,\
     unit_weapon_type_animation_names, vehicle_animation_names,\
     weapon_animation_names, device_animation_names, fp_animation_names,\
     unit_damage_animation_names
from reclaimer.hek.defs.mod2 import TagDef, Pad, marker as mod2_marker_desc,\
     node as mod2_node_desc, reflexive, Struct

__all__ = (
    'SHARED_UNIT_ANIMATION_NAMES', 'SHARED_UNIT_WEAPON_ANIMATION_NAMES',
    'split_anim_name_into_type_strings', 'split_permutation_number',
    'set_animation_enum_index', 'set_animation_index',
    'get_default_animation_enums', 'set_default_animation_enums',
    'partial_mod2_def', 'calculate_node_vectors'
    )


SHARED_UNIT_ANIMATION_NAMES = frozenset((
    'airborne-dead', 'landing-dead', 'talk', 'emotions'
    ))

SHARED_UNIT_WEAPON_ANIMATION_NAMES = frozenset((
    'dive-front', 'dive-back', 'dive-left', 'dive-right', 'airborne',
    'land-soft', 'land-hard', 'throw-grenade', 'berserk',
    'surprise-front', 'surprise-back', 'evade-left', 'evade-right',
    'signal-move', 'signal-attack', 'warn', 'melee', 'celebrate', 'panic',
    'melee-airborne', 'flaming', 'resurrect-front', 'resurrect-back',
    'melee-continuous', 'feeding', 'leap-start', 'leap-airborne', 'leap-melee'
    ))


partial_mod2_def = TagDef("mod2",
    Pad(64),
    Struct('tagdata',
        Pad(172),
        reflexive("markers", mod2_marker_desc, 256, DYN_NAME_PATH=".name"),
        reflexive("nodes", mod2_node_desc, 64, DYN_NAME_PATH=".name"),
        SIZE=232
        ),
    ext=".gbxmodel", endian=">"
    )


def split_anim_name_into_type_strings(anim_name):
    anim_name_len = -1
    while len(anim_name) != anim_name_len:
        anim_name_len = len(anim_name)
        anim_name = anim_name.replace("  ", " ")

    name_pieces = anim_name.split(" ", 4)
    part1 = name_pieces[0] if len(name_pieces) > 0 else ""
    part2 = name_pieces[1] if len(name_pieces) > 1 else ""
    part3 = name_pieces[2] if len(name_pieces) > 2 else ""
    part4 = name_pieces[3] if len(name_pieces) > 3 else ""

    part1_sani = part1.lower().replace("_", "-")
    part2_sani = part2.lower().replace("_", "-")
    part3_sani = part3.lower().replace("_", "-")
    part4_sani = part4.lower().replace("_", "-")

    remainder = " ".join(name_pieces[4: ])

    type_strings = ()
    perm_num = ""

    if part1_sani == "suspension":
        if part4: remainder = " ".join((part4, remainder))
        if part3: remainder = " ".join((part3, remainder))
        if part2: remainder = " ".join((part2, remainder))

        remainder, _, perm_num = split_permutation_number(remainder)
        type_strings = part1_sani, remainder.lower(), perm_num
        remainder = ""

    elif part1_sani in ("first-person", "device", "vehicle"):
        if part4: remainder = " ".join((part4, remainder))
        if part3: remainder = " ".join((part3, remainder))

        part2, part2_sani, perm_num = split_permutation_number(part2, remainder)
        if ((part1_sani == "first-person" and part2_sani in fp_animation_names) or
            (part1_sani == "vehicle" and part2_sani in vehicle_animation_names) or
            (part1_sani == "device"  and part2_sani in device_animation_names)):
            type_strings = part1_sani, part2_sani, perm_num

    elif (part1_sani in ("s-ping", "h-ping", "s-kill", "h-kill") and
          part2_sani in ("front", "left", "right", "back")):

        if part4: remainder = " ".join((part4, remainder))

        part3, part3_sani, perm_num = split_permutation_number(part3, remainder)
        if part3_sani in ("gut", "chest", "head", "l-arm", "l-hand", "l-leg",
                          "l-foot", "r-arm", "r-hand", "r-leg", "r-foot"):
            type_strings = part1_sani, part2_sani, part3_sani, perm_num

    else:
        part4, part4_sani, perm_num = split_permutation_number(part4, remainder)
        if part4_sani in unit_weapon_type_animation_names:
            type_strings = part1, part2, part3, part4_sani, perm_num

        else:
            part3, part3_sani, perm_num = split_permutation_number(part3, remainder)
            if part3_sani in unit_weapon_animation_names:
                if part4: remainder = " ".join((part4, remainder))

                type_strings = part1, part2, part3_sani, perm_num

            else:
                part2, part2_sani, perm_num = split_permutation_number(part2, remainder)
                if part2_sani in unit_animation_names:
                    if part4: remainder = " ".join((part4, remainder))
                    if part3: remainder = " ".join((part3, remainder))

                    type_strings = part1, part2_sani, perm_num
                else:
                    remainder = True

    # nothing should have a remainder. if it does, it doesnt fit the criteria,
    # and we should just return it split at the permutation character
    if remainder:
        name_pieces = anim_name.lower().split("%")
        if len(name_pieces) > 1:
            anim_name = "%".join(name_pieces[: -1])
            perm_num  = name_pieces[-1]

        return False, (anim_name, perm_num)

    return True, type_strings


def split_permutation_number(*strings):
    string = " ".join(s for s in strings if s)
    pieces = string.split("%")
    part0 = pieces[0]
    sani_part0 = part0.lower().replace("_", "-")
    if len(pieces) > 1:
        return part0, sani_part0, pieces[1].lstrip()
    return part0, sani_part0, ""


def set_animation_enum_index(anim_enums, enum_index, anim_index,
                             indices_to_not_overwrite):
    if enum_index >= len(anim_enums):
        anim_enums.extend(enum_index + 1 - len(anim_enums))

    if anim_enums[enum_index].animation in indices_to_not_overwrite:
        return False
    anim_enums[enum_index].animation = anim_index
    return True


def get_default_animation_enums(anim_enums, defaults):
    for i in defaults:
        if i not in range(len(anim_enums)):
            continue
        elif anim_enums[i].animation >= 0 and defaults[i] < 0:
            # this animation is set, and we DON'T have a valid default.
            # set the default to this valid animation enum
            defaults[i] = anim_enums[i].animation


def set_default_animation_enums(anim_enums, defaults):
    for i in defaults:
        if i not in range(len(anim_enums)):
            continue
        elif anim_enums[i].animation < 0 and defaults[i] >= 0:
            # this animation is unset, and we DO have a valid default.
            # set this animation enum to the valid default
            anim_enums[i].animation = defaults[i]


def set_animation_index(antr_tag, anim_name, anim_index,
                        indices_to_not_overwrite=()):
    orig_anim_name = anim_name

    tagdata = antr_tag.data.tagdata
    antr_units    = tagdata.units.STEPTREE
    antr_weapons  = tagdata.weapons.STEPTREE
    antr_vehicles = tagdata.vehicles.STEPTREE
    antr_devices  = tagdata.devices.STEPTREE
    antr_unit_damages  = tagdata.unit_damages.STEPTREE
    antr_fp_animations = tagdata.fp_animations.STEPTREE
    antr_anims = tagdata.animations.STEPTREE
    antr_nodes = tagdata.nodes.STEPTREE

    _, name_pieces = split_anim_name_into_type_strings(anim_name)
    name_pieces = list(name_pieces)
    perm_num = name_pieces.pop(-1)

    part1 = name_pieces[0] if len(name_pieces) > 0 else ""
    part2 = name_pieces[1] if len(name_pieces) > 1 else ""
    part3 = name_pieces[2] if len(name_pieces) > 2 else ""
    part4 = name_pieces[3] if len(name_pieces) > 3 else ""

    if part1 in ("vehicle", "suspension") and len(antr_vehicles) < 1:
        # default right/left yaw/pitch to 1 frame count
        # and yaw and pitch to 60 each
        antr_vehicles.append()
        block = antr_vehicles[0]
        block.right_frame_count = block.left_frame_count = 1
        block.down_frame_count  = block.up_frame_count   = 1
        block.right_yaw_per_frame  = block.left_yaw_per_frame = pi / 3
        block.down_pitch_per_frame = block.up_pitch_per_frame = pi / 3

    if part1 == "suspension":
        anim_name = anim_name.split(" ", 1)[1]
        anim_enums = antr_vehicles[0].suspension_animations.STEPTREE
        if len(anim_enums) >= 8:
            # max of 8 suspension animations
            return False

        anim_enums.append()
        anim_enums[-1].animation = anim_index
        return True

    elif part1 in ("first-person", "device", "vehicle"):
        if part1 == "first-person":
            options = fp_animation_names
            block = antr_fp_animations
        elif part1 == "device":
            options = device_animation_names
            block = antr_devices
        elif part1 == "vehicle":
            options = vehicle_animation_names
            block = antr_vehicles

        if not block:
            block.append()

        try:
            enum_index = options.index(part2)
        except ValueError:
            return False

        return set_animation_enum_index(block[0].animations.STEPTREE, enum_index,
                                        anim_index, indices_to_not_overwrite)

    elif part1 in ("s-ping", "h-ping", "s-kill", "h-kill"):
        # divided into 16 chunks of 11 sets of animations
        #   NOTE: Defaults to "gut" if no matches can be found
        #         Otherwise, defaults to the first matching
        #         region. Ex:  l-hand will default to chest if
        #                      l-hand and l-arm are missing
        #  0  ==  gut                       1  ==  chest
        #  2  ==  head / chest              3  ==  l-arm / chest
        #  4  ==  l-hand / l-arm / chest    5  ==  l-leg
        #  6  ==  l-foot / l-leg            7  ==  r-arm / chest
        #  8  ==  r-hand / l-arm / chest    9  ==  r-leg
        # 10  ==  r-foot / r-leg
        #
        # this means that the unit-damages should be set in this order:
        #     r-foot, r-leg, r-hand, r-arm,
        #     l-foot, l-leg, l-hand, l-arm,
        #     head, chest, gut
        #
        # and make sure to set lowest number permutation first
        #
        # each of these sections of 44 is divided into 4
        # sections of 11 animations, each in this order:
        #     front, left, right, back
        #
        #   0 -  43  ==  s-ping
        #  44 -  87  ==  h-ping
        #  88 - 131  ==  s-kill
        # 132 - 175  ==  h-kill
        #
        # so for example, "s-kill right l-foot" would be
        # 88 + 11*2 + 6   which is index 116

        try:
            enum_index = unit_damage_animation_names.index(
                " ".join((part1, part2, part3))
                )
            base = (enum_index // 11) * 11
            enum_index = enum_index % 11
        except ValueError:
            return False

        if len(antr_unit_damages) < 176:
            antr_unit_damages.extend(176 - len(antr_unit_damages))

        indices_to_set = range(1, 11)
        if   enum_index == 1: indices_to_set = [2, 3, 4, 7, 8]
        elif enum_index == 3: indices_to_set = [4]
        elif enum_index == 5: indices_to_set = [6]
        elif enum_index == 7: indices_to_set = [8]
        elif enum_index == 9: indices_to_set = [10]

        if not set_animation_enum_index(antr_unit_damages, base + enum_index,
                                        anim_index, indices_to_not_overwrite):
            return False

        # set any that havent been set to anything
        for i in indices_to_set:
            block = antr_unit_damages[base + i]
            if block.animation < 0:
                block.animation = anim_index

        return True

    elif part1 in weapon_animation_names:
        enum_index = weapon_animation_names.index(part1)
        if len(antr_weapons) < 1:
            antr_weapons.append()

        return set_animation_enum_index(antr_weapons[0].animations.STEPTREE,
                                        enum_index, anim_index,
                                        indices_to_not_overwrite)

    # something in the units
    if part2 in unit_animation_names:
        typ = 1
        enum_index = unit_animation_names.index(part2)
    elif part3 in unit_weapon_animation_names:
        typ = 2
        enum_index = unit_weapon_animation_names.index(part3)
    elif part4 in unit_weapon_type_animation_names:
        typ = 3
        enum_index = unit_weapon_type_animation_names.index(part4)
    else:
        return False

    unit = None
    for block in antr_units:
        if block.label == part1:
            unit = block
            break

    if unit is None:
        if len(antr_units) >= antr_units.MAX:
            return False

        antr_units.append()
        unit = antr_units[-1]

        unit.label = part1
        # default right/left yaw/pitch to 1 frame count
        # and yaw and pitch to 60 each
        unit.right_frame_count = unit.left_frame_count = 1
        unit.down_frame_count  = unit.up_frame_count   = 1
        unit.right_yaw_per_frame  = unit.left_yaw_per_frame = pi / 3
        unit.down_pitch_per_frame = unit.up_pitch_per_frame = pi / 3

        # tool likes to have the first 12 always exist, even if unset
        unit.animations.STEPTREE.extend(12 - len(unit.animations.STEPTREE))

    if typ == 1:
        # unit animation
        return set_animation_enum_index(unit.animations.STEPTREE, enum_index,
                                        anim_index, indices_to_not_overwrite)


    unit_weaps = unit.weapons.STEPTREE
    unit_weap = None
    for block in unit_weaps:
        if block.name == part2:
            unit_weap = block
            break

    if unit_weap is None:
        if len(unit_weaps) >= unit_weaps.MAX:
            return False

        unit_weaps.append()
        unit_weap = unit_weaps[-1]

        unit_weap.name = part2
        # default right/left yaw/pitch to 1 frame count
        # and yaw and pitch to 60 each
        unit_weap.right_frame_count = unit_weap.left_frame_count = 1
        unit_weap.down_frame_count  = unit_weap.up_frame_count   = 1
        unit_weap.right_yaw_per_frame  = unit_weap.left_yaw_per_frame = pi / 3
        unit_weap.down_pitch_per_frame = unit_weap.up_pitch_per_frame = pi / 3

    if typ == 2:
        # unit weapon animation
        return set_animation_enum_index(unit_weap.animations.STEPTREE,
                                        enum_index, anim_index,
                                        indices_to_not_overwrite)

    unit_weap_types = unit_weap.weapon_types.STEPTREE
    unit_weap_type = None
    for block in unit_weap_types:
        if block.label == part3:
            unit_weap_type = block
            break

    if unit_weap_type is None:
        if len(unit_weap_types) >= unit_weap_types.MAX:
            return False

        unit_weap_types.append()
        unit_weap_type = unit_weap_types[-1]
        unit_weap_type.label = part3

    # unit weapon type animation
    return set_animation_enum_index(unit_weap_type.animations.STEPTREE,
                                    enum_index, anim_index,
                                    indices_to_not_overwrite)


def calculate_node_vectors(antr_nodes, mod2_nodes, jma_anims):
    # NOTE: The base vector is a ray starting at the node that points
    # relative to the parent node. It points toward the center of the
    # bounds of all the rotations of that node across all animations.
    # The vector range is twice the max rotation in any direction that
    # node will vary across all animations.
    # ball-socket type means the joint can pitch, yaw, and roll freely.
    # hinge type means the joint can pitch and roll freely.

    # The cyborgs head has a base vector of 80 degrees, points up, and
    # the joint is set as a ball socket. This means the head can pitch,
    # yaw, and roll 40 degrees in any direction from the base vector.
    pass



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# DO NOT REMOVE! __all__ is used to locate frozen definitions
__all__ = ("composer_playlist", "pc_gametype", "xbox_gametype")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Veegie
# 	Mapped out reflextives and idents. Fixed titles.
# revision: 3		author: DarkShallFall
# 	Mapped out some values in barrel, new triggers, etc.
# revision: 4		author: DarkShallFall
# 	Mapped out and labeled all of the "Firing Effect" reflexive.
# revision: 5		author: DeadCanadian
# 	fixed third person flag, and named sounds
# revision: 6		author: Lord Zedd
# 	Portin'
# revision: 7		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

weap_barrel_angle_change_function = (
    "linear",
    "early",
    "very_early",
    "late",
    "very_late",
    "cosine",
    "one",
    "zero",
    )

weap_barrel_distribution_function = (
    "point",
    "horizontal_fan",
    )

weap_barrel_prediction_type = (
    "none",
    "continuous",
    "instant",
    )

weap_movement_penalized = (
    "always",
    "when_zoomed",
    "when_zoomed_or_reloading",
    )

weap_multiplayer_weapon_type = (
    "none",
    "ctf_flag",
    "oddball_ball",
    "headhunter_head",
    "juggernaut_powerup",
    )

weap_new_trigger_behavior = (
    "spew",
    "latch",
    "latch_autofire",
    "charge",
    "latch_zoom",
    "latch_rocketlauncher",
    "spew_charge",
    )

weap_new_trigger_button_used = (
    "right_trigger",
    "left_trigger",
    "melee_attack",
    "automated_fire",
    "right_bumper",
    )

weap_new_trigger_overcharge_action = (
    "none",
    "explode",
    "discharge",
    )

weap_new_trigger_prediction = (
    "none",
    "spew",
    "charge",
    )

weap_new_trigger_secondary_action = (
    "fire",
    "charge",
    "track",
    "fire_other",
    )

weap_secondary_trigger_mode = (
    "normal",
    "slaved_to_primary",
    "inhibits_primary",
    "loads_alternate_ammunition",
    "loads_multiple_primary_ammunition",
    )

weap_tracking_type = (
    "no_tracking",
    "human_tracking",
    "plasma_tracking",
    )

weap_weapon_type = (
    "undefined",
    "shotgun",
    "needler",
    "plasma_pistol",
    "plasma_rifle",
    "rocket_launcher",
    "energy_sword",
    "spartan_laser",
    )


weap_early_mover_propertie = Struct("early_mover_propertie",
    h3_string_id("name"),
    Pad(36),
    ENDIAN=">", SIZE=40
    )


weap_ai_propertie = Struct("ai_propertie",
    Bool32("flags",
        "destroyable_cover",
        "pathfinding_ignore_when_dead",
        "dynamic_cover",
        ),
    h3_string_id("ai_type_name"),
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    SEnum16("size", *bloc_ai_propertie_size),
    SEnum16("leap_jump_speed", *bloc_ai_propertie_leap_jump_speed),
    ENDIAN=">", SIZE=16
    )


weap_function = Struct("function",
    Bool32("flags",
        "invert",
        "mapping_does_not_controls_active",
        "always_active",
        "random_time_offset",
        ),
    h3_string_id("import_name"),
    h3_string_id("export_name"),
    h3_string_id("turn_off_with"),
    Float("minimum_value"),
    h3_rawdata_ref("default_function"),
    h3_string_id("scale_by"),
    ENDIAN=">", SIZE=44
    )


weap_attachment = Struct("attachment",
    h3_dependency("attachment"),
    h3_string_id("marker"),
    SEnum16("change_color", *bloc_attachment_change_color),
    SInt16("unknown", VISIBLE=False),
    h3_string_id("primary_scale"),
    h3_string_id("secondary_scale"),
    ENDIAN=">", SIZE=32
    )


weap_widget = Struct("widget",
    h3_dependency("type"),
    ENDIAN=">", SIZE=16
    )


weap_change_color_initial_permutation = Struct("initial_permutation",
    Pad(4),
    color_rgb_float("color_lower_bound"),
    color_rgb_float("color_upper_bound"),
    h3_string_id("variant_name"),
    ENDIAN=">", SIZE=32
    )


weap_change_color_function = Struct("function",
    Bool32("scale_flags",
        "blend_in_hsv",
        "more_colors",
        ),
    color_rgb_float("color_lower_bound"),
    color_rgb_float("color_upper_bound"),
    h3_string_id("darken_by"),
    h3_string_id("scale_by"),
    ENDIAN=">", SIZE=32
    )


weap_change_color = Struct("change_color",
    h3_reflexive("initial_permutations", weap_change_color_initial_permutation),
    h3_reflexive("functions", weap_change_color_function),
    ENDIAN=">", SIZE=24
    )


weap_predicted_resources_0 = Struct("predicted_resources_0",
    SInt16("type"),
    SInt16("resource_index"),
    dependency_uint32("tag_index", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=8
    )


weap_multiplayer_object_propertie = Struct("multiplayer_object_propertie",
    Bool16("engine_flags",
        "capture_the_flag",
        "slayer",
        "oddball",
        "king_of_the_hill",
        "juggernaut",
        "territories",
        "assault",
        "vip",
        "infection",
        ),
    SEnum8("object_type", *bloc_multiplayer_object_propertie_object_type),
    Bool8("teleporter_flags",
        "disallows_players",
        "allows_land_vehicles",
        "allows_heavy_vehicles",
        "allows_flying_vehicles",
        "allows_projectiles",
        ),
    Bool16("flags",
        "editor_only",
        ),
    SEnum8("shape", *bloc_multiplayer_object_propertie_shape),
    SEnum8("spawn_timer_mode", *bloc_multiplayer_object_propertie_spawn_timer_mode),
    SInt16("spawn_time"),
    SInt16("abandon_time"),
    Float("radius_width"),
    Float("length"),
    Float("top"),
    Float("bottom"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    SInt32("unknown_3", VISIBLE=False),
    SInt32("unknown_4", VISIBLE=False),
    h3_dependency("child_object"),
    SInt32("unknown_5"),
    h3_dependency("shape_shader"),
    h3_dependency("unknown_shader"),
    h3_dependency("unknown_6"),
    h3_dependency("unknown_7"),
    h3_dependency("unknown_8"),
    h3_dependency("unknown_9"),
    h3_dependency("unknown_10"),
    h3_dependency("unknown_11"),
    ENDIAN=">", SIZE=196
    )


weap_predicted_bitmap = Struct("predicted_bitmap",
    h3_dependency("bitmap", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


weap_first_person = Struct("first_person",
    h3_dependency("first_person_model"),
    h3_dependency("first_person_animations"),
    ENDIAN=">", SIZE=32
    )


weap_predicted_resources_1 = Struct("predicted_resources_1",
    SInt16("type"),
    SInt16("resource_index"),
    dependency_uint32("tag_index", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=8
    )


weap_magazine_magazine_equipment = Struct("magazine_equipment",
    SInt16("rounds_0_for_max"),
    SInt16("unknown", VISIBLE=False),
    h3_dependency("equipment"),
    ENDIAN=">", SIZE=20
    )


weap_magazine = Struct("magazine",
    Bool32("flags",
        "wastes_rounds_when_reloaded",
        "every_round_must_be_chambered",
        ),
    SInt16("rounds_recharged"),
    SInt16("rounds_total_initial"),
    SInt16("rounds_total_maximum"),
    SInt16("rounds_total_loaded_maximum"),
    SInt16("maximum_rounds_held"),
    SInt16("unknown_0"),
    Float("reload_time"),
    SInt16("rounds_reloaded"),
    SInt16("unknown_1", VISIBLE=False),
    Float("chamber_time"),
    BytesRaw("unknown_2", SIZE=24, VISIBLE=False),
    h3_dependency("reloading_effect"),
    h3_dependency("reloading_damage_effect"),
    h3_dependency("chambering_effect"),
    h3_dependency("chambering_damage_effect"),
    h3_reflexive("magazine_equipment", weap_magazine_magazine_equipment),
    ENDIAN=">", SIZE=128
    )


weap_new_trigger = Struct("new_trigger",
    Bool32("flags",
        "autofire_single_action_only",
        ),
    SEnum16("button_used", *weap_new_trigger_button_used),
    SEnum16("behavior", *weap_new_trigger_behavior),
    SInt16("primary_barrel"),
    SInt16("secondary_barrel"),
    SEnum16("prediction", *weap_new_trigger_prediction),
    SInt16("unknown_0", VISIBLE=False),
    Float("autofire_time"),
    Float("autofire_throw"),
    SEnum16("secondary_action", *weap_new_trigger_secondary_action),
    SEnum16("primary_action", *weap_new_trigger_secondary_action),
    Float("charging_time"),
    Float("charged_time"),
    SEnum16("overcharge_action", *weap_new_trigger_overcharge_action),
    Bool16("charge_flags", *unknown_flags_16),
    Float("charged_illumination"),
    Float("spew_time"),
    h3_dependency("charging_effect"),
    h3_dependency("charging_damage_effect"),
    h3_dependency("charging_response"),
    Float("charging_age_degeneration"),
    h3_dependency("discharging_effect"),
    h3_dependency("discharging_damage_effect"),
    Float("target_tracking_decay_time"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=144
    )


weap_barrel_first_person_offset = Struct("first_person_offset",
    QStruct("first_person_offset", INCLUDE=xyz_float),
    ENDIAN=">", SIZE=12
    )


weap_barrel_firing_effect = Struct("firing_effect",
    SInt16("shot_count_lower_bound"),
    SInt16("shot_count_upper_bound"),
    h3_dependency("firing_effect"),
    h3_dependency("misfire_effect"),
    h3_dependency("empty_effect"),
    h3_dependency("unknown_effect"),
    h3_dependency("firing_response"),
    h3_dependency("misfire_response"),
    h3_dependency("empty_response"),
    h3_dependency("unknown_response"),
    h3_dependency("rider_firing_response"),
    h3_dependency("rider_misfire_response"),
    h3_dependency("rider_empty_response"),
    h3_dependency("rider_unknown_response"),
    ENDIAN=">", SIZE=196
    )


weap_barrel = Struct("barrel",
    Bool32("flags",
        "tracks_fired_projectile",
        "random_firing_effects",
        "can_fire_with_partial_ammo",
        "projectiles_use_weapon_origin",
        "ejects_during_chamber",
        "use_error_when_unzoomed",
        "projectile_vector_cannot_be_adjusted",
        "projectiles_have_identical_error",
        "projectiles_fire_parallel",
        "cant_fire_when_others_firing",
        "cant_fire_when_others_recovering",
        "dont_clear_fire_bit_after_recovering",
        "stagger_fire_across_multiple_markers",
        "fires_locked_projectiles",
        ),
    QStruct("rounds_per_second", INCLUDE=from_to),
    Float("acceleration_time_0"),
    Float("deceleration_time_0"),
    Float("barrel_spin_scale"),
    Float("blurred_rate_of_fire"),
    QStruct("shots_per_fire", INCLUDE=from_to_sint16),
    Float("fire_recovery_time"),
    Float("soft_recovery_fraction"),
    SInt16("magazine"),
    SInt16("rounds_per_shot"),
    SInt16("minimum_rounds_loaded"),
    SInt16("rounds_between_tracers"),
    h3_string_id("optional_barrel_marker_name"),
    SEnum16("prediction_type", *weap_barrel_prediction_type),
    SEnum16("firing_noise", *gint_constant_sound_volume),
    Float("acceleration_time_1"),
    Float("deceleration_time_1"),
    QStruct("damage_error", INCLUDE=from_to),
    float_rad("base_turning_speed"),
    from_to_rad("dynamic_turning_speed"),
    Float("acceleration_time_2"),
    Float("deceleration_time_2"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    float_rad("minimum_error_0"),
    from_to_rad("error_angle_0"),
    Float("dual_wield_damage_scale"),
    SEnum16("distribution_function", *weap_barrel_distribution_function),
    SInt16("projectiles_per_shot"),
    float_rad("distribution_angle"),
    float_rad("minimum_error_1"),
    from_to_rad("error_angle_1"),
    h3_reflexive("first_person_offsets", weap_barrel_first_person_offset),
    SEnum8("damage_reporting_type", *proj_damage_reporting_type),
    SInt8("unknown_2", VISIBLE=False),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("initial_projectile"),
    h3_dependency("trailing_projectile"),
    h3_dependency("damage_effect"),
    h3_dependency("crate_projectile"),
    Float("crate_projectile_speed"),
    Float("ejection_port_recovery_time"),
    Float("illumination_recovery_time"),
    Float("heat_generated_per_round"),
    Float("age_generated_per_round_mp"),
    Float("age_generated_per_round_sp"),
    Float("overload_time"),
    from_to_rad("angle_change_per_shot"),
    Pad(8),
    SEnum16("angle_change_function", *weap_barrel_angle_change_function),
    SInt16("unknown_4", VISIBLE=False),
    Float("unknown_5", VISIBLE=False),
    Float("unknown_6", VISIBLE=False),
    Float("firing_effect_deceleration_time"),
    Float("unknown_7", VISIBLE=False),
    Float("rate_of_fire_acceleration_time"),
    Float("rate_of_fire_deceleration_time"),
    Float("unknown_8", VISIBLE=False),
    Float("unknown_9", VISIBLE=False),
    h3_reflexive("firing_effects", weap_barrel_firing_effect),
    ENDIAN=">", SIZE=308
    )


weap_body = Struct("tagdata",
    SEnum16("object_type", *bloc_object_type),
    Bool16("flags_0",
        "does_not_cast_shadow",
        "search_cardinal_direction_lightmaps",
        ("not_a_pathfinding_obstacle", 1 << 3),
        "extension_of_parent",
        "does_not_cause_collision_damage",
        "early_mover",
        "early_mover_localized_physics",
        "use_static_massive_lightmap_sample",
        "object_scales_attachments",
        "inherits_player_s_appearance",
        "dead_bipeds_can_t_localize",
        "attach_to_clusters_by_dynamic_sphere",
        "effects_created_by_this_object_do_not_spawn_objects_in_multiplayer",
        ),
    Float("bounding_radius"),
    QStruct("bounding_offset", INCLUDE=xyz_float),
    Float("acceleration_scale"),
    SEnum16("lightmap_shadow_mode_size", *bloc_lightmap_shadow_mode_size),
    SEnum8("sweetener_size", *bloc_sweetener_size),
    SEnum8("water_density", *bloc_water_density),
    SInt32("unknown_0", VISIBLE=False),
    Float("dynamic_light_sphere_radius"),
    QStruct("dynamic_light_sphere_offset", INCLUDE=xyz_float),
    h3_string_id("default_model_variant"),
    h3_dependency("model"),
    h3_dependency("crate_object"),
    h3_dependency("collision_damage"),
    h3_reflexive("early_mover_properties", weap_early_mover_propertie),
    h3_dependency("creation_effect"),
    h3_dependency("material_effects"),
    h3_dependency("melee_impact"),
    h3_reflexive("ai_properties", weap_ai_propertie),
    h3_reflexive("functions", weap_function),
    SInt16("hud_text_message_index"),
    SInt16("unknown_1", VISIBLE=False),
    h3_reflexive("attachments", weap_attachment),
    h3_reflexive("widgets", weap_widget),
    h3_reflexive("change_colors", weap_change_color),
    h3_reflexive("predicted_resources_0", weap_predicted_resources_0),
    h3_reflexive("multiplayer_object_properties", weap_multiplayer_object_propertie),
    Bool32("flags_1",
        "always_maintains_z_up",
        "destroyed_by_explosions",
        "unaffected_by_gravity",
        ),
    SInt16("old_message_index"),
    SInt16("sort_order"),
    Float("old_multiplayer_on_ground_scale"),
    Float("old_campaign_on_ground_scale"),
    h3_string_id("pickup_message"),
    h3_string_id("swap_message"),
    h3_string_id("pickup_or_dual_wield_message"),
    h3_string_id("swap_or_dual_wield_message"),
    h3_string_id("picked_up_message"),
    h3_string_id("switch_to_message"),
    h3_string_id("switch_to_from_ai_message"),
    h3_string_id("all_weapons_empty_message"),
    h3_dependency("collision_sound"),
    h3_reflexive("predicted_bitmaps", weap_predicted_bitmap),
    h3_dependency("detonation_damage_effect_0"),
    QStruct("detonation_delay", INCLUDE=from_to),
    h3_dependency("detonating_effect"),
    h3_dependency("detonation_effect"),
    Float("campaign_ground_scale"),
    Float("multiplayer_ground_scale"),
    Float("small_hold_scale"),
    Float("small_holster_scale"),
    Float("medium_hold_scale"),
    Float("medium_holster_scale"),
    Float("large_hold_scale"),
    Float("large_holster_scale"),
    Float("huge_hold_scale"),
    Float("huge_holster_scale"),
    Float("grounded_friction_length"),
    Float("grounded_friction_unknown"),
    Bool32("flags_2",
        "vertical_heat_display",
        "mutually_exclusive_triggers",
        "attacks_automatically_on_bump",
        "must_be_readied",
        "doesn_t_count_towards_maximum",
        "aim_assists_only_when_zoomed",
        "prevents_grenade_throwing",
        "must_be_picked_up",
        "holds_triggers_when_dropped",
        "prevents_melee_attack",
        "detonates_when_dropped",
        "cannot_fire_at_maximum_age",
        "secondary_trigger_overrides_grenades",
        "does_not_depower_active_camo_in_multiplayer",
        "enables_integrated_night_vision",
        "ais_use_weapon_melee_damage",
        "forces_no_binoculars",
        "loop_fp_firing_animation",
        "prevents_sprinting",
        "cannot_fire_while_boosting",
        "prevents_driving",
        "third_person_camera",
        "can_be_dual_wielded",
        "can_only_be_dual_wielded",
        "melee_only",
        "can_t_fire_if_parent_dead",
        "weapon_ages_with_each_kill",
        "weapon_uses_old_dual_fire_error_code",
        "primary_trigger_melee_attacks",
        "cannot_be_used_by_player",
        ),
    Bool32("more_flags", *unknown_flags_32),
    h3_string_id("unknown_2"),
    SEnum16("secondary_trigger_mode", *weap_secondary_trigger_mode),
    SInt16("maximum_alternate_shots_loaded"),
    Float("turn_on_time"),
    Float("ready_time"),
    h3_dependency("ready_effect"),
    h3_dependency("ready_damage_effect"),
    Float("heat_recovery_threshold"),
    Float("overheated_threshold"),
    Float("heat_detonation_threshold"),
    Float("heat_detonation_fraction"),
    Float("heat_loss_per_second"),
    Float("heat_illumination"),
    Float("overheated_heat_loss_per_second"),
    h3_dependency("overheated"),
    h3_dependency("overheated_damage_effect"),
    h3_dependency("detonation"),
    h3_dependency("detonation_damage_effect_1"),
    h3_dependency("player_melee_damage"),
    h3_dependency("player_melee_response"),
    yp_float_rad("damage_pyramid_angles"),
    Float("damage_pyramid_depth"),
    h3_dependency("_1st_hit_damage"),
    h3_dependency("_1st_hit_response"),
    h3_dependency("_2nd_hit_damage"),
    h3_dependency("_2nd_hit_response"),
    h3_dependency("_3rd_hit_damage"),
    h3_dependency("_3rd_hit_response"),
    h3_dependency("lunge_melee_damage"),
    h3_dependency("lunge_melee_response"),
    h3_dependency("gun_gun_clang_damage"),
    h3_dependency("gun_gun_clang_response"),
    h3_dependency("gun_sword_clang_damage"),
    h3_dependency("gun_sword_clang_response"),
    h3_dependency("clash_effect"),
    SEnum8("melee_damage_reporting_type", *proj_damage_reporting_type),
    SInt8("unknown_3", VISIBLE=False),
    SInt16("magnification_levels"),
    QStruct("magnification_range", INCLUDE=from_to),
    float_rad("autoaim_angle"),
    Float("autoaim_range_long"),
    Float("autoaim_range_short"),
    float_rad("magnetism_angle"),
    Float("magnetism_range_long"),
    Float("magnetism_range_short"),
    float_rad("deviation_angle"),
    BytesRaw("unknown_4", SIZE=20, VISIBLE=False),
    SEnum16("movement_penalized", *weap_movement_penalized),
    SInt16("unknown_5", VISIBLE=False),
    Float("forwards_movement_penalty"),
    Float("sideways_movement_penalty"),
    Float("ai_scariness"),
    Float("weapon_power_on_time"),
    Float("weapon_power_off_time"),
    h3_dependency("weapon_power_on_effect"),
    h3_dependency("weapon_power_off_effect"),
    Float("age_heat_recovery_penalty"),
    Float("age_rate_of_fire_penalty"),
    Float("age_misfire_start"),
    Float("age_misfire_chance"),
    h3_dependency("pickup_sound"),
    h3_dependency("zoom_in_sound"),
    h3_dependency("zoom_out_sound"),
    Float("active_camo_ding"),
    Float("active_camo_regrowth_rate"),
    h3_string_id("handle_node"),
    h3_string_id("weapon_class"),
    h3_string_id("weapon_name"),
    SEnum16("multiplayer_weapon_type", *weap_multiplayer_weapon_type),
    SEnum16("weapon_type", *weap_weapon_type),
    SEnum16("tracking_type", *weap_tracking_type),
    SInt16("unknown_6", VISIBLE=False),
    BytesRaw("unknown_7", SIZE=16, VISIBLE=False),
    h3_reflexive("first_person", weap_first_person),
    h3_dependency("hud_interface"),
    h3_reflexive("predicted_resources_1", weap_predicted_resources_1),
    h3_reflexive("magazines", weap_magazine),
    h3_reflexive("new_triggers", weap_new_trigger),
    h3_reflexive("barrels", weap_barrel),
    Float("unknown_8", VISIBLE=False),
    Float("unknown_9", VISIBLE=False),
    Float("maximum_movement_acceleration"),
    Float("maximum_movement_velocity"),
    Float("maximum_turning_acceleration"),
    Float("maximum_turning_velocity"),
    h3_dependency("deployed_vehicle"),
    h3_dependency("deployed_weapon"),
    h3_dependency("age_model"),
    h3_dependency("age_weapon"),
    h3_dependency("aged_effect"),
    Float("hammer_age_per_use_mp"),
    Float("hammer_age_per_use_sp"),
    QStruct("first_person_weapon_offset", INCLUDE=ijk_float),
    QStruct("first_person_scope_size", INCLUDE=ij_float),
    QStruct("third_person_pitch_bounds", INCLUDE=from_to),
    Float("zoom_transition_time"),
    Float("melee_weapon_delay"),
    h3_string_id("weapon_holster_marker"),
    ENDIAN=">", SIZE=1280
    )


def get():
    return weap_def

weap_def = TagDef("weap",
    h3_blam_header('weap'),
    weap_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["weap"], endian=">", tag_cls=H3Tag
    )
